import gc
import sys
from time import sleep

from naeural_core import Logger

if __name__ == '__main__':
  
  l = Logger('MEM', base_folder='.', app_folder='_local_cache')
  
  class A:
    def __init__(self, obj=None):
      self.a = 'a' * 100
      self.obj = obj
      
  class B:
    def __init__(self, obj=None):
      self.b = 'b' * 100
      self.obj = obj


  class C:
    def __init__(self, obj=None):
      self.b = {
        k : [1] *k for k in range(1000)
      }
      self.obj = obj
      
  a = A()
  b = B(obj=a)
    
  sa = l.get_obj_size(a)

  
  c = C()
  sc, tree = l.get_obj_size(c, return_tree=True)
  
  for _ in range(1000):
    l.start_timer('get_obj_size')
    sc, tree = l.get_obj_size(c, return_tree=True)
    l.stop_timer('get_obj_size')
  l.show_timers()
  
  
  # import torch as th
  # import numpy as np
  
  # gc.collect()
  # mem1 = l.get_avail_memory(False)  
  # l.P('{:0,.0f}'.format(mem1 / 1024**2))

  # tc = th.tensor(np.ones((2000,1000,1000)), dtype=th.float32)

  # gc.collect()
  # sleep(1)
  # mem2 = l.get_avail_memory(False)
  # l.P('{:0,.0f}: {}'.format(mem2 / 1024**2, sys.getsizeof(tc)))

  # tc = tc.to(th.device('cuda'))

  # gc.collect()
  # sleep(1)
  # mem3 = l.get_avail_memory(False)
  # l.P('{:0,.0f}: {}'.format(mem3 / 1024**2, sys.getsizeof(tc)))

  # tc = tc.type(th.uint8)

  # gc.collect()
  # sleep(1)
  # mem4 = l.get_avail_memory(False)
  # l.P('{:0,.0f}: {}'.format(mem4 / 1024**2, sys.getsizeof(tc)))
  
  
  # tcf = th.tensor(np.ones((2000,1000,1000)), dtype=th.float32)

  # gc.collect()
  # mem5 = l.get_avail_memory(False)
  # l.P('{:0,.0f}'.format(mem5 / 1024**2))


  # tcf = tcf.to(th.device('cuda'))

  # gc.collect()
  # mem6 = l.get_avail_memory(False)
  # l.P('{:0,.0f}'.format(mem6 / 1024**2))


import numpy as np
import torch as th
import torchvision as tv

def box_iou(box1, box2):
  # https://github.com/pytorch/vision/blob/master/torchvision/ops/boxes.py
  """
  Return intersection-over-union (Jaccard index) of boxes.
  Both sets of boxes are expected to be in (x1, y1, x2, y2) format.
  Arguments:
      box1 (Tensor[N, 4])
      box2 (Tensor[M, 4])
  Returns:
      iou (Tensor[N, M]): the NxM matrix containing the pairwise
          IoU values for every element in boxes1 and boxes2
  """

  def box_area(box):
    # box = 4xn
    return (box[2] - box[0]) * (box[3] - box[1])

  area1 = box_area(box1.T)
  area2 = box_area(box2.T)

  # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
  inter = (th.min(box1[:, None, 2:], box2[:, 2:]) - th.max(box1[:, None, :2], box2[:, :2])).clamp(0).prod(2)
  return inter / (area1[:, None] + area2 - inter)  # iou = inter / (area1 + area2 - inter)


def xywh2xyxy(x):
  """
  Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the
  top-left corner and (x2, y2) is the bottom-right corner.

  Args:
      x (np.ndarray) or (torch.Tensor): The input bounding box coordinates in (x, y, width, height) format.
  Returns:
      y (np.ndarray) or (torch.Tensor): The bounding box coordinates in (x1, y1, x2, y2) format.
  """
  y = x.clone()
  y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
  y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
  y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
  y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
  return y

@th.jit.script
def y5_nms_topk(prediction,
           # conf_thres:float=0.25,
           # iou_thres:float=0.45,
           # classes=None,
           # agnostic=False,
           # multi_label:bool=False,
           # max_det:int=300
           ):
  """Runs Non-Maximum Suppression (NMS) on inference results

  Returns:
        list of detections, on (n,6) tensor per image [xyxy, conf, cls]
  """

  nc = prediction.shape[2] - 5  # number of classes
  xc = prediction[..., 4] > 0.25  # candidates #conf_thres
  # Settings
  max_wh = 7680  # (pixels) minimum and maximum box width and height
  max_nms = 30000  # maximum number of boxes into tv.ops.nms()
  n_candidates = 6
  bs = prediction.shape[0]

  # output = [th.zeros((0, 6), device=prediction.device)] * 256 # max batch size 256
  th_output = th.zeros((bs, 300, 4 + 2 * n_candidates), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)
  for xi in range(bs):  # image index, image inference
    x = prediction[xi]
    x = x[xc[xi]]  # confidence

    # If none remain process next image
    if x.shape[0] > 0:
      # Compute conf
      x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

      # Box (center x, center y, width, height) to (x1, y1, x2, y2)
      box = xywh2xyxy(x[:, :4])

      # Detections matrix nx(4+2*n_candidates) (xyxy, conf1, cls1, .., confi, clsi)
      # conf, j = x[:, 5:].max(1, keepdim=True) # confidente & class-id
      # x = th.cat((box, conf, j.float()), 1)[conf.view(-1) > 0.25] #conf_thres

      top_conf, top_idxs = x[:, 5:].sort(1, descending=True)
      mask = top_conf[:, 0] > 0.25
      runners = th.stack([top_conf[:, :n_candidates], top_idxs[:, :n_candidates]], dim=2).view(top_conf.shape[0],
                                                                                                   2 * n_candidates)
      x = th.cat((box, runners), 1)[mask]

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence

        # Batched NMS
        c = x[:, 5:6] * max_wh  # classes preparred as offsets
        boxes = x[:, :4] + c # add offeset to put each class in "different scene"
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300] #max_det

        th_n_det[xi] = i.shape[0]
        th_output[xi, :th_n_det[xi], :] = x[i]

  return th_output, th_n_det


@th.jit.script
def y5_nms(
    prediction,
   # conf_thres:float=0.25,
   # iou_thres:float=0.45,
   # classes=None,
   # agnostic=False,
   # multi_label:bool=False,
   # max_det:int=300
):
  """Runs Non-Maximum Suppression (NMS) on inference results

    Returns:
          list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """

  nc = prediction.shape[2] - 5  # number of classes
  xc = prediction[..., 4] > 0.25  # candidates #conf_thres
  # Settings
  max_wh = 7680  # (pixels) minimum and maximum box width and height
  max_nms = 30000  # maximum number of boxes into tv.ops.nms()
  bs = prediction.shape[0]

  # output = [th.zeros((0, 6), device=prediction.device)] * 256 # max batch size 256
  th_output = th.zeros((bs, 300, 6), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)
  for xi in range(bs):  # image index, image inference
    x = prediction[xi]
    x = x[xc[xi]]  # confidence

    # If none remain process next image
    if x.shape[0] > 0:
      # Compute conf
      x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

      # Box (center x, center y, width, height) to (x1, y1, x2, y2)
      box = xywh2xyxy(x[:, :4])

      # Detections matrix nx(4+2*n_candidates) (xyxy, conf1, cls1, .., confi, clsi)
      conf, j = x[:, 5:].max(1, keepdim=True) # confidente & class-id
      x = th.cat((box, conf, j.float()), 1)[conf.view(-1) > 0.25] #conf_thres

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence

        # Batched NMS
        c = x[:, 5:6] * max_wh  # classes preparred as offsets
        boxes = x[:, :4] + c  # add offeset to put each class in "different scene"
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300]  # max_det

        th_n_det[xi] = i.shape[0]
        th_output[xi, :th_n_det[xi], :] = x[i]

  return th_output, th_n_det


@th.jit.script
def y8_nms_topk(
        prediction,
        # conf_thres:float=0.25,
        # iou_thres:float=0.45,
        # classes=None,
        # agnostic=False,
        # multi_label:bool=False,
        # max_det:int=300,
        # max_time_img:float=0.05,
        # max_nms:int=30000,
        # max_wh:int=7680,
):
  """
  Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

  Arguments:
      prediction (th.Tensor): A tensor of shape (batch_size, num_boxes, num_classes + 4 + num_masks)
          containing the predicted boxes, classes, and masks. The tensor should be in the format
          output by a model, such as YOLO.
      conf_thres (float): The confidence threshold below which boxes will be filtered out.
          Valid values are between 0.0 and 1.0.
      iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
          Valid values are between 0.0 and 1.0.
      classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
      agnostic (bool): If True, the model is agnostic to the number of classes, and all
          classes will be considered as one.
      multi_label (bool): If True, each box may have multiple labels.
      labels (List[List[Union[int, float, th.Tensor]]]): A list of lists, where each inner
          list contains the apriori labels for a given image. The list should be in the format
          output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
      max_det (int): The maximum number of boxes to keep after NMS.
      nc (int): (optional) The number of classes output by the model. Any indices after this will be considered masks.
      max_time_img (float): The maximum time (seconds) for processing one image.
      max_nms (int): The maximum number of boxes into tv.ops.nms().
      max_wh (int): The maximum box width and height in pixels

  Returns:
      (List[th.Tensor]): A list of length batch_size, where each element is a tensor of
          shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
          (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
  """

  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nm = 0 # prediction.shape[1] - nc - 4
  mi = 4 + nc  # mask start index
  xc = prediction[:, 4:mi].amax(1) > 0.25  # candidates  #conf_thres
  n_candidates = 6

  th_output = th.zeros((bs, 300, 4 + 2 * n_candidates + nm), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)

  for xi in range(prediction.shape[0]):  # image index, image inference
    x = prediction[xi]
    x = x.transpose(0, -1)[xc[xi]]  # confidence

    # If none remain process next image
    if x.shape[0] > 0:

      # Detections matrix nx6 (xyxy, conf, cls)
      box, cls, mask = x.split((4, nc, nm), 1)
      box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)

      # conf, j = cls.max(1, keepdim=True)
      # x = th.cat((box, conf, j.float(), mask), 1)

      top_conf, top_idxs = x[:, 4:].sort(1, descending=True)
      runners = th.stack([top_conf[:, :n_candidates], top_idxs[:, :n_candidates]], dim=2)
      runners = runners.view(top_conf.shape[0], 2 * n_candidates)
      x = th.cat((box, runners, mask), 1)

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:30000]]  # sort by confidence and remove excess boxes #max_nms

        # Batched NMS
        c = x[:, 5:6] * 7680  # classes #max_wh
        boxes = x[:, :4] + c
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300]  # limit detections #max_det

        th_n_det[xi] = i.shape[0]
        th_output[xi, :th_n_det[xi], :] = x[i]

  return th_output, th_n_det


@th.jit.script
def y8_nms(
    prediction,
    # conf_thres:float=0.25,
    # iou_thres:float=0.45,
    # classes=None,
    # agnostic=False,
    # multi_label:bool=False,
    # max_det:int=300,
    # max_time_img:float=0.05,
    # max_nms:int=30000,
    # max_wh:int=7680,
):
  """
  Perform non-maximum suppression (NMS) on a set of boxes, with support for masks and multiple labels per box.

  Arguments:
      prediction (th.Tensor): A tensor of shape (batch_size, num_boxes, num_classes + 4 + num_masks)
          containing the predicted boxes, classes, and masks. The tensor should be in the format
          output by a model, such as YOLO.
      conf_thres (float): The confidence threshold below which boxes will be filtered out.
          Valid values are between 0.0 and 1.0.
      iou_thres (float): The IoU threshold below which boxes will be filtered out during NMS.
          Valid values are between 0.0 and 1.0.
      classes (List[int]): A list of class indices to consider. If None, all classes will be considered.
      agnostic (bool): If True, the model is agnostic to the number of classes, and all
          classes will be considered as one.
      multi_label (bool): If True, each box may have multiple labels.
      labels (List[List[Union[int, float, th.Tensor]]]): A list of lists, where each inner
          list contains the apriori labels for a given image. The list should be in the format
          output by a dataloader, with each label being a tuple of (class_index, x1, y1, x2, y2).
      max_det (int): The maximum number of boxes to keep after NMS.
      nc (int): (optional) The number of classes output by the model. Any indices after this will be considered masks.
      max_time_img (float): The maximum time (seconds) for processing one image.
      max_nms (int): The maximum number of boxes into tv.ops.nms().
      max_wh (int): The maximum box width and height in pixels

  Returns:
      (List[th.Tensor]): A list of length batch_size, where each element is a tensor of
          shape (num_boxes, 6 + num_masks) containing the kept boxes, with columns
          (x1, y1, x2, y2, confidence, class, mask1, mask2, ...).
  """

  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nm = 0  # prediction.shape[1] - nc - 4
  mi = 4 + nc  # mask start index
  xc = prediction[:, 4:mi].amax(1) > 0.25  # candidates  #conf_thres

  th_output = th.zeros((bs, 300, 6 + nm), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)

  for xi in range(prediction.shape[0]):  # image index, image inference
    x = prediction[xi]
    x = x.transpose(0, -1)[xc[xi]]  # confidence

    # If none remain process next image
    if x.shape[0] > 0:

      # Detections matrix nx6 (xyxy, conf, cls)
      box, cls, mask = x.split((4, nc, nm), 1)
      box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)

      conf, j = cls.max(1, keepdim=True)
      x = th.cat((box, conf, j.float(), mask), 1)

      # Check shape
      n = x.shape[0]  # number of boxes
      if n > 0:  # no boxes
        x = x[x[:, 4].argsort(descending=True)[:30000]]  # sort by confidence and remove excess boxes #max_nms

        # Batched NMS
        c = x[:, 5:6] * 7680  # classes #max_wh
        boxes = x[:, :4] + c
        scores = x[:, 4]  # boxes (offset by class), scores
        i = tv.ops.nms(boxes, scores, 0.45)  # NMS #iou_thres
        i = i[:300]  # limit detections #max_det

        th_n_det[xi] = i.shape[0]
        th_output[xi, :th_n_det[xi], :] = x[i]

  return th_output, th_n_det

# See https://github.com/NVIDIA/TensorRT/tree/32c64a324e58f252eae4e5681f5c39dbe22ef2d5/plugin/efficientNMSPlugin
class TRT_EfficientNMS(th.autograd.Function):
  """
  An autograd.Function that produces an invocation of the
  TensorRT EfficientNMS plugin operation in its graph. The torch
  results are invalid and should only be used for tracing.
  """

  @staticmethod
  def forward(
    ctx,
    boxes,
    scores,
    background_class=-1,
    box_coding=0,
    iou_threshold=0.45,
    max_output_boxes=300,
    plugin_version="1",
    score_activation=0,
    score_threshold=0.25,
  ):
    # Only required to produce data of the right type to continue tracing.
    # This will not be part of the ONNX graph.
    batch_size, _, num_classes = scores.shape
    num_det = th.randint(0, max_output_boxes, (batch_size, 1), dtype=th.int32)
    det_boxes = th.randn(batch_size, max_output_boxes, 4)
    det_scores = th.randn(batch_size, max_output_boxes)
    det_classes = th.randint(0, num_classes, (batch_size, max_output_boxes), dtype=th.int32)
    return num_det, det_boxes, det_scores, det_classes

  # See https://pytorch.org/docs/stable/onnx_torchscript.html#static-symbolic-method
  @staticmethod
  def symbolic(
    g : th.Graph, # The ONNX graph
    boxes,
    scores,
    background_class=-1,
    box_coding=0,
    iou_threshold=0.45,
    max_output_boxes=300,
    plugin_version="1",
    score_activation=0,
    score_threshold=0.25
  ):
    # The args have _i, _f, _s suffixes to indicate the type of the argument.
    # boxes and scores are positional arguments so there is no need for _i, _f, _s.
    det_nums, det_boxes, det_scores, det_classes = g.op(
      "TRT::EfficientNMS_TRT",
      boxes,
      scores,
      background_class_i=background_class,
      box_coding_i=box_coding,
      iou_threshold_f=iou_threshold,
      max_output_boxes_i=max_output_boxes,
      plugin_version_s=plugin_version,
      score_activation_i=score_activation,
      score_threshold_f=score_threshold,
      class_agnostic_i=0,
      outputs=4
    )

    # This is not a node that ONNX knows about and we need to
    # manually set the output shapes.
    # bs x 1, int32
    det_nums.setType(det_nums.type().with_dtype(th.int32).with_sizes([None, 1]))
    # bs x max_output_boxes x 4
    det_boxes.setType(det_boxes.type().with_sizes([None, 300, 4]))
    # bs x max_output_boxes
    det_scores.setType(det_scores.type().with_sizes([None, 300]))
    # bs x max_output_boxes, int32
    det_classes.setType(det_classes.type().with_dtype(th.int32).with_sizes([None, 300]))

    return det_nums, det_boxes, det_scores, det_classes

def y8_nms_trt_efficient(prediction : th.Tensor):
  """
  A NMS implementation compatible with TensorRT which can be exported to an ONNX
  graph. The IOU threshold used is 0.45 and the score threshold is 0.25.
  Invoking this in torch will not produce correct results and should only
  be used for tracing purposes.

  Arguments:
    prediction (th.Tensor): A tensor of shape (batch_size, num_classes + 4, num_boxes)
      containing the predicted boxes, classes, and masks. The tensor should be in the format
      output by a model, such as YOLO.

  Returns:
    Tensor containg the detections. The tensor has shape batch_size x 300 x 6.
      Each input image will have a maximum of 300 detections, each detection
      being in the format x1, y1, x2, y2, confidence, class.

    Tensor containing the number of detections for each image in the batch. The
      tensor has a shape of batch_size x 1.
  """
  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nb = prediction.shape[2]
  nm = 0  # prediction.shape[1] - nc - 4
  mi = 4 + nc  # mask start index

  # Shuffle input data to match the TRT NMS expected format.
  th_output = th.zeros((bs, 300, 6 + nm), device=prediction.device)
  th_n_det = th.zeros(bs, device=prediction.device, dtype=th.int32)
  prediction = prediction.transpose(1, -1)  # confidence

  box, cls, _ = prediction.split((4, nc, nm), 2)
  box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)

  # The original y8_nms will suppress non maximum classes so we do
  # the same here. Set confidences to zero for all classes not selected
  # by tensor.max. Note this will select only one class, even if we have
  # two classes with the same confidence.
  # Not sure I understand the point of this operation. It might be
  # perfectly sensible to have more than one class per box, so this might be
  # a limiation of the original implementation.
  # The below commented code would be faster but not the same in edge cases.
  # Code:
  #    th_max, _ = cls.max(2, keepdim=True)
  #    cls = cls * cls.eq(th_max).to(cls.dtype)
  # The fastest alternative is of course to not do this at all.
  _, indices = cls.max(2, keepdim=True)
  max_mask = th.zeros(cls.shape, dtype=cls.dtype, device=prediction.device)
  cls = cls * max_mask.scatter_(2, indices, 1.)

  # Run the TRT Efficient NMS plugin.
  num_detections, nmsed_boxes, nmsed_scores, nmsed_classes = TRT_EfficientNMS.apply(
    box, # boxes
    cls, # scores
    -1, # background_class
    0, # box_coding, boxes in xyxy format
    0.45, # iou_threshold
    300, # max_output_boxes
    '1', # plugin_version
    0, # score_activation
    0.25 # score_threshold
  )

  # Use th.view as a way to tell shape inference about the shape of the output.
  num_detections = num_detections.view(bs, 1)
  nmsed_scores = nmsed_scores.view(bs, 300, 1)
  nmsed_classes = nmsed_classes.view(bs, 300, 1)
  nmsed_boxes = nmsed_boxes.view(bs, 300, 4)

  # Reorder the TRT NMS output data to our expected NMS output format.
  th_output[:, :nmsed_boxes.shape[1], :4] = nmsed_boxes     # xyxy
  th_output[:, :nmsed_scores.shape[1], 4:5] = nmsed_scores  # confidence
  th_output[:, :nmsed_scores.shape[1], 5:6] = nmsed_classes # class

  # Once more, make sure shape inference can properly see the output shape.
  th_n_det = num_detections.view(bs)
  th_output = th_output.view(bs, 300, 6)

  return th_output, th_n_det

class ONNX_NMS(th.autograd.Function):
  @staticmethod
  def forward(
    ctx,
    boxes,
    scores,
    max_output_boxes_per_class,
    iou_threshold,
    score_threshold
  ):
    import random
    batch_size, num_classes, num_boxes = scores.shape

    all_dets = []
    for _ in range(batch_size):
      num_det = random.randint(2, 5)

      det_batch_idx = th.randint(0, batch_size, (num_det, 1), device=boxes.device)
      det_class_idx = th.randint(0, num_classes, (num_det, 1), device=boxes.device)
      det_box_idx = th.randint(0, num_boxes, (num_det, 1), device=boxes.device)

      all_dets.append(th.cat((det_batch_idx, det_class_idx, det_box_idx), 1))
    return th.cat(all_dets, 0).to(th.int64)

  @staticmethod
  def symbolic(
    g : th._C.Graph, # The ONNX graph
    boxes,
    scores,
    max_output_boxes_per_class,
    iou_threshold,
    score_threshold
  ):
    out = g.op(
      "NonMaxSuppression",
      boxes,
      scores,
      max_output_boxes_per_class,
      iou_threshold,
      score_threshold,
      center_point_box_i=0, #xyxy
      outputs=1
    )
    return out


def bin_count(th_x, batch_size):
  # An implementation of a one dimensional integer bincount.
  # This is used because th.bincount will stop at the max value
  # seen in th_x.
  i = th.arange(0, batch_size, 1, dtype=th.int32, device=th_x.device)
  th_idx = i.view(-1, 1).repeat(1, th_x.shape[0])
  return (th_idx[i, :] == th_x).sum(dim=1).view(batch_size, 1)


def y8_nms_onnx(prediction):
  device=prediction.device
  bs = prediction.shape[0]  # batch size
  nc = prediction.shape[1] - 4  # number of classes
  nb = prediction.shape[2]
  nm = 0  # prediction.shape[1] - nc - 4

  # FIXME: y8_nms would remove boxes with according to the
  # confidence threshold here. We don't really need to do this
  # as the ONNX nms does have a confidence threshold.

  prediction = prediction.transpose(1, -1)  # confidence

  th_output = th.zeros((bs, 300, 6 + nm), device=device)

  box, cls, _ = prediction.split((4, nc, nm), 2)
  box = xywh2xyxy(box)  # center_x, center_y, width, height) to (x1, y1, x2, y2)

  conf, j = cls.max(2, keepdim=True)
  x = th.cat((box, conf, j.float()), 2)

  c = x[:, :, 5:6] * 7680  # classes #max_wh
  boxes = x[:, :, :4] + c
  scores = x[:, :, 4]  # boxes (offset by class), scores
  scores = scores.view(bs, 1, nb) # bs X 1 X nb

   # We're using shared classes so the number of classes in boxes_in is 1.
  boxes = boxes.view(bs, nb, 4)

  # Note this does NOT have a constant shape.
  indices = ONNX_NMS.apply(
    boxes, # boxes,
    scores, # scores
    th.tensor([300], dtype=th.int64, device=device), # max_output_boxes_per_class
    th.tensor([0.45], dtype=th.float32, device=device), # iou_threshold
    th.tensor([0.25], dtype=th.float32, device=device) # score_threshold
  )

  # For every detection left we get the initial batch index and the box list index.
  batch_idx, box_idx = indices[:, 0], indices[:, 2]

  # Sort by batch indices. Looking at the ONNXRT code
  # this might not be required (at least not on the CPU side).
  sorted_batch_idx = th.argsort(batch_idx, dim=0)
  batch_idx = batch_idx[sorted_batch_idx]
  box_idx = box_idx[sorted_batch_idx]

  # We can use bin_count to get the number of detections per each batch.
  # th.bincount would not work here because it stops at the last value
  # for which there is a count.
  th_n_det = bin_count(batch_idx, bs)
  end_idx = th.cumsum(th_n_det, dim=0, dtype=th.int32)
  start_idx = end_idx - th_n_det

  batch_idx = batch_idx.view(indices.shape[0], 1)
  box_idx = box_idx.view(indices.shape[0], 1)
  boxes = x[batch_idx, box_idx].view(indices.shape[0], 6)

  # Construct a 1d scatter mask to get the output in the format
  # that we expect. There may be an easier way to do this but I
  # have no idea what that would be.
  # We need to use a scatter and cannot index because ONNX falls
  # over with indexing.
  batch_offsets = th.arange(0, batch_idx.shape[0], 1, device=device)
  batch_offsets = batch_offsets.view(-1, 1) - start_idx[batch_idx].view(-1, 1)
  batch_offsets = batch_offsets + batch_idx * 300

  # Change the mask so we can pick up all elements.
  batch_offsets = batch_offsets.repeat((1, 6))
  batch_offsets = (6 * batch_offsets).view(-1,6) + th.arange(0, 6, 1, dtype=th.int32, device=device).view(-1, 6)

  # Change all tensors to 1d and do the scatter.
  batch_offsets = batch_offsets.view(-1)
  th_output = th_output.view(-1)
  boxes = boxes.view(-1)
  th_output = th_output.scatter_(dim=0, index=batch_offsets, src=boxes)

  # This helps with the onnx shape inference.
  th_output = th_output.view(bs, 300, 6)

  # Sort outputs by confidence. Only needed for OpenVINO, however this triggers
  # an assert... Maybe it's time for scatter again?
  # This is not needed for now, since the sorting of the detections does not affect
  # the plugins that use this output
  # batch_idx = th.arange(bs).repeat_interleave(300).view(bs * 300)
  # conf_idx = th.argsort(th_output[:,:,4], dim=1, descending=True).view(bs * 300)
  # The following 2 lines produce the same output, but neither will work for ONNX.
  # th_output = th_output[batch_idx, conf_idx, :].view(bs, 300, 6)
  # th_output = th_output.view(bs * 300, 6)[batch_offsets].view(bs, 300, 6)

  return th_output, th_n_det.view(-1)



from functools import partial

from ratio1 import Logger as BaseLogger
from .logger_mixins import (
  _AdvancedTFKerasMixin,
  _BasicPyTorchMixin,
  _BasicTFKerasMixin,
  _BetaInferenceMixin,
  _ComplexNumpyOperationsMixin,
  _ConfusionMatrixMixin,
  _DataFrameMixin,
  _DeployModelsInProductionMixin,
  _FitDebugTFKerasMixin,
  _GPUMixin,
  _GridSearchMixin,
  _HistogramMixin,
  _KerasCallbacksMixin,
  _MatplotlibMixin,
  _MultithreadingMixin,
  _NLPMixin,
  _PackageLoaderMixin,
  _PublicTFKerasMixin,
  _TF2ModulesMixin,
  _TimeseriesBenchmakerMixin,
  _VectorSpaceMixin,
  _LlmUtilsMixin
)


class Logger(
  BaseLogger,
  _MatplotlibMixin,
  _ConfusionMatrixMixin,
  _HistogramMixin,
  _ComplexNumpyOperationsMixin,
  _DataFrameMixin,
  _NLPMixin,
  _GridSearchMixin,
  _TimeseriesBenchmakerMixin,
  _VectorSpaceMixin,
  _LlmUtilsMixin,
  _GPUMixin,
  _PackageLoaderMixin,
  _PublicTFKerasMixin,
  _BasicTFKerasMixin,
  _AdvancedTFKerasMixin,
  _DeployModelsInProductionMixin,
  _FitDebugTFKerasMixin,
  _KerasCallbacksMixin,
  _TF2ModulesMixin,
  _BasicPyTorchMixin,
  _BetaInferenceMixin,
  _MultithreadingMixin,
):

  def __init__(self, lib_name="",
               lib_ver="",
               config_file="",
               base_folder=None,
               app_folder=None,
               config_data={},
               show_time=True,
               config_file_encoding=None,
               no_folders_no_save=False,
               max_lines=None,
               HTML=False,
               DEBUG=True,
               data_config_subfolder=None,
               check_additional_configs=False,
               TF_KERAS=False,
               BENCHMARKER=False,
               dct_bp_params=None,
               default_color='n',
               ):

    super(Logger, self).__init__(
      lib_name=lib_name, lib_ver=lib_ver,
      config_file=config_file,
      base_folder=base_folder,
      app_folder=app_folder,
      show_time=show_time,
      config_data=config_data,
      config_file_encoding=config_file_encoding,
      no_folders_no_save=no_folders_no_save,
      max_lines=max_lines,
      HTML=HTML,
      DEBUG=DEBUG,
      data_config_subfolder=data_config_subfolder,
      check_additional_configs=check_additional_configs,
      default_color=default_color,
    )

    how_runs = ''
    if self.runs_from_ipython():
      how_runs = ' running in ipython'
    if self.runs_with_debugger():
      how_runs = ' running in debug mode'
    self.verbose_log(
        "  Python {}{}, Git branch: '{}', Conda: '{}'".format(
          self.python_version,
          how_runs,
          self.git_branch,
          self.conda_env,
        ), color='green'
      )

    self.verbose_log("  Os: {}".format(self.get_os_name()), color='g')
    self.verbose_log("  IP: {}".format(self.get_localhost_ip()), color='g')

    self.verbose_log('  Avail/Total RAM: {:.1f} GB / {:.1f} GB'.format(
      self.get_avail_memory(), self.get_machine_memory()
    ), color='green')

    if BENCHMARKER:
      dct_bp_params = dct_bp_params or {}
      self._setup_benchmarker(**dct_bp_params)

    if TF_KERAS:
      self.check_tf()

    return


SBLogger = partial(Logger, lib_name='tst', base_folder='.', app_folder='_local_cache')

if __name__ == '__main__':
  l = Logger('TEST', base_folder='Dropbox', app_folder='_libraries_testdata')
  l.P("All check", color='green')



import inspect
import os
import torch as th
import pandas as pd
import numpy as np
import traceback
import abc
from functools import partial
from typing import Any, Dict
from time import time
from naeural_core import DecentrAIObject
from naeural_core import Logger
from naeural_core.local_libraries.nn.th.image_dataset_stage_preprocesser import preprocess_to_str
# TODO !!!! move this to non-experimental location
from naeural_core.utils.tracing.base_torch_scripter import BaseTorchScripter
from naeural_core.local_libraries.nn.th.trainer import ModelTrainer
from ratio1 import _PluginsManagerMixin
from naeural_core import constants as ct
from collections import defaultdict


class BaseTrainingPipeline(DecentrAIObject, _PluginsManagerMixin):
  score_key = None
  score_mode = None

  data_factory_loc = None
  model_factory_loc = None
  training_callbacks_loc = None

  def __init__(
      self, log: Logger, signature: str, config: dict,
      path_to_dataset: str, dataset_disk_preprocess_kwargs: dict = None, **kwargs
  ):
    """
    Base class for training pipelines.
    Parameters
    ----------
    log : Logger, logger object
    signature : str, signature of the pipeline
    config : dict, configuration dictionary
    path_to_dataset : str, path to the dataset
    dataset_disk_preprocess_kwargs : dict, dictionary with disk preprocess
    This will be used in order to preprocess the dataset before loading it.
    kwargs : dict, additional keyword arguments
    """
    self.config = config
    self.signature = signature

    self._dct_data_factories = {'train' : None, 'dev' : None, 'test' : None}
    self._dct_data_dirs = {'train' : '', 'dev' : '', 'test' : ''}

    self._path_to_dataset = path_to_dataset
    self._dataset_name = os.path.split(self._path_to_dataset.rstrip(os.sep))[-1]# .split(os.sep)[-1]
    self.dataset_disk_preprocess_kwargs = dataset_disk_preprocess_kwargs

    self._data_factory_ref = None
    self._model_factory_ref = None
    self._training_callbacks_ref = None

    self._trainer = None
    self._trainer_callbacks_obj = None

    self._status_per_grid_idx = {}
    self._top_k_best_model_idx = None
    self._time_start_run = None
    self._grid_loop_exception = False

    self.df_model_name_to_hiperparams : pd.DataFrame = None

    self.export_done = False
    self._status : Dict[str, Any] = {
      'ELAPSED' : None,
      'REMAINING' : None,
      'GRID_ITER' : None,
      'NR_ALL_GRID_ITER' : None,
      'GRID_RESULTS' : None,
      'BEST' : None,
    }

    self._metadata : Dict[str, Any] = {
      'CLASSES' : None,
      'MODEL_ARCHITECTURE_PATH' : None,
      'MODEL_NAME' : None,
      'FIRST_STAGE_TARGET_CLASS' : None,
      'INFERENCE_PREPROCESS_DEFINITIONS' : None,
      'MODEL_EXPORT': None
    }
    super(BaseTrainingPipeline, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    def _append_if_not_void(lst, x):
      if x is not None:
        lst.append(x)
      return

    super().startup()

    if self.score_key is None or self.score_mode is None:
      raise ValueError("score_key/score_mode") # TODO

    self._find_data_directories(self._path_to_dataset)

    if len(self._dct_data_dirs['train']) == 0 or len(self._dct_data_dirs['dev']) == 0:
      raise ValueError("train/dev not existing") # TODO

    data_factory_loc = ['local_libraries.nn.th.training.data']
    _append_if_not_void(data_factory_loc, self.data_factory_loc)

    model_factory_loc = ['local_libraries.nn.th.training.models']
    _append_if_not_void(model_factory_loc, self.model_factory_loc)

    training_callbacks_loc = ['local_libraries.nn.th.training.callbacks']
    _append_if_not_void(training_callbacks_loc, self.training_callbacks_loc)

    self._data_factory_ref = self._get_module_name_and_class(
      locations=data_factory_loc,
      name=self.signature,
      search_in_packages=ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES,
      suffix='DataLoaderFactory'
    )

    self._model_factory_ref = self._get_module_name_and_class(
      locations=model_factory_loc,
      name=self.signature,
      search_in_packages=ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES,
      suffix='ModelFactory'
    )

    self._training_callbacks_ref = self._get_module_name_and_class(
      locations=training_callbacks_loc,
      name=self.signature,
      search_in_packages=ct.PLUGIN_SEARCH.SEARCH_IN_PACKAGES,
      suffix='TrainingCallbacks'
    )

    self._metadata['MODEL_ARCHITECTURE_PATH'] = '{}.{}'.format(
      self._model_factory_ref[0].__name__,
      self._model_factory_ref[1],
    )

    self._metadata['MODEL_NAME'] = self.cfg_model_name
    self._metadata['FIRST_STAGE_TARGET_CLASS'] = self.cfg_first_stage_target_class

    self.experiment_subfolder = 'training_' + self.log.session_id + '_{}'.format(self.cfg_model_name)

    self.experiment_models_folder = os.path.join(
      self.log.get_models_folder(),
      self.experiment_subfolder
    )

    return

  def disk_preprocessed_name(self):
    """
    Method that will be called in order to get the name of the preprocessed dataset.
    Returns
    -------
    res : str, name of the preprocessed dataset
    """
    preprocess_params = self.dataset_disk_preprocess_kwargs
    params_values = [preprocess_params.get(k, '') for k in sorted(preprocess_params.keys())]
    return f"{self._dataset_name}__{'__'.join([str(val) if val is not None else '' for val in params_values])}"

  def disk_preprocess_datapoint(self, datapoint_files, src_path_prefix, dst_path_prefix):
    """
    Method that will be called in order to preprocess a datapoint before loading it.
    When implementing this in a child class, the user will need to define preprocessing
    of a datapoint, but will also need to define the saving of the new datapoint in the
    new dataset location.
    Parameters
    ----------
    datapoint_files : list, list of files that are part of the datapoint
    src_path_prefix : str, path prefix to the source dataset
    dst_path_prefix : str, path prefix to the destination dataset
    """
    raise NotImplementedError

  def disk_preprocess_dataset_init(self, **kwargs):
    """
    Utility method that will be called before the preprocessing itself,
    in order for the user to be able to properly set up the disk preprocessing of the dataset.
    """
    pass

  def disk_preprocess_dataset(self):
    """
    Method that will be called in order to preprocess the dataset before loading it.
    The preprocessed dataset will be saved on disk with the name provided by the `disk_preprocessed_name` method.
    Returns
    -------

    """
    # First, the setup of the disk preprocessing will be done
    self.disk_preprocess_dataset_init(**{
      k.lower(): v for k, v in self.dataset_disk_preprocess_kwargs.items()
    })
    # The new dataset directory needs to be created
    new_dataset_name = self.disk_preprocessed_name()
    ds_path_prefix = os.path.split(self._path_to_dataset.rstrip(os.sep))[0]
    full_new_ds_path = os.path.join(ds_path_prefix, new_dataset_name)
    if os.path.exists(full_new_ds_path):
      # The dataset already exists, so we will not preprocess it again
      self.P(f"Dataset {new_dataset_name} already exists. Skipping disk preprocessing.")
      return full_new_ds_path, new_dataset_name
    # endif dataset already exists

    # The dataset will be preprocessed
    # Firstly, a partial directory will be created
    partial_dataset_name = f'{new_dataset_name}_partial'
    ds_path = os.path.join(ds_path_prefix, partial_dataset_name)
    os.makedirs(ds_path, exist_ok=True)

    # For each file basename all the extensions will be gathered.
    # This is done in order to have access to all the information of a certain datapoint,
    # since the dataset can have multiple files for each datapoint.
    files_dict = defaultdict(lambda: [])
    for root, dirs, files in os.walk(self._path_to_dataset):
      for file in files:
        basename, ext = os.path.splitext(file)
        files_dict[basename].append(ext)
      # endfor files
    # endfor os.walk

    # All the datapoints will be preprocessed
    for basename, exts in files_dict.items():
      datapoint_files = [f'{basename}.{ext}' for ext in exts]
      self.disk_preprocess_datapoint(datapoint_files, self._path_to_dataset, ds_path)
    # endfor datapoints
    # The partial dataset will be renamed to the final dataset name
    os.rename(ds_path, full_new_ds_path)
    return full_new_ds_path, new_dataset_name

  def maybe_disk_preprocess_dataset(self):
    """
    Method that will be called before the training starts.
    It will check if the dataset needs to be preprocessed before loading it.
    """
    if self.dataset_disk_preprocess_kwargs is not None:
      self.P("Disk preprocessing dataset...")
      new_ds_path, new_ds_name = self.disk_preprocess_dataset()
      self._path_to_dataset = new_ds_path
      self._dataset_name = new_ds_name
      self.P(f"Dataset preprocessed. New dataset path: {new_ds_path} and new dataset name: {new_ds_name}. Performing startup again...")
      self.startup()
    return

  @abc.abstractmethod
  def model_loss(self, **dct_grid_option):
    """
    Defines the model loss. Grid parameters can be used.

    Parameters:
    ----------
    **dct_grid_option
      Current grid option

    Returns:
    -------
    loss_obj
      A loss object that can be called with 2 parameters (y_pred, y) and returns the final loss.
      Most of the times, torch losses will be used (e.g. th.nn.CrossEntropyLoss() )
    """
    raise NotImplementedError

  @property
  def cfg_batch_index(self):
    return self.config.get('BATCH_INDEX', True)

  @property
  def cfg_num_workers(self):
    return self.config.get("NUM_WORKERS", None)

  @property
  def cfg_model_name(self):
    return self.config['MODEL_NAME']

  @property
  def cfg_preload_data(self):
    return self.config.get('PRELOAD_DATA', True)

  @property
  def cfg_device_load_data(self):
    return self.config.get('DEVICE_LOAD_DATA', 'cuda:0')

  @property
  def cfg_device_training(self):
    return self.config.get('DEVICE_TRAINING', 'cuda:0')

  @property
  def cfg_start_grid_idx(self):
    return self.config.get('START_GRID_IDX', None)

  @property
  def cfg_end_grid_idx(self):
    return self.config.get('END_GRID_IDX', None)

  @property
  def cfg_grid_search(self):
    """
    Property that lets the pipeline to use other grid search, not the one defined in the architecture.
    """
    return self.config.get('GRID_SEARCH', {})

  def get_grid_search(self):
    """
    Getter for the grid search dictionary.
    This is used in order to be able to override it in a child class.
    Returns
    -------
    dict
    """
    return self.cfg_grid_search

  @property
  def cfg_batch_size(self):
    return self.config['BATCH_SIZE']

  @property
  def cfg_epochs(self):
    return self.config['EPOCHS']

  @property
  def cfg_first_stage_target_class(self):
    return self.config.get('FIRST_STAGE_TARGET_CLASS', None)

  @property
  def cfg_keep_top_k_iterations(self):
    """
    Property that tells the pipeline to save only the best K models in a grid search in order to save disk space.
    """
    return self.config.get('KEEP_TOP_K_ITERATIONS', None)

  @property
  def cfg_pretrained_weights_url(self):
    """
    Property that returns the URL from where the pretrained weights can be downloaded or the path to the pretrained weights.
    """
    return self.config.get('PRETRAINED_WEIGHTS_URL', None)

  @property
  def cfg_export_model(self):
    """
    Property that tells the pipeline to export the model after training.
    """
    return self.config.get('EXPORT_MODEL', False)

  @property
  def default_grid_search(self):
    dct = self._model_factory_ref[3] or {}
    return dct.get('GRID_SEARCH', {})

  @property
  def model_factory_class_def(self):
    return self._model_factory_ref[2]

  @property
  def training_callbacks_class_def(self):
    return self._training_callbacks_ref[2]

  @property
  def status(self):
    if self._time_start_run is not None:
      self._status['ELAPSED'] = time() - self._time_start_run

    if self.df_model_name_to_hiperparams is not None:
      self._status['GRID_RESULTS'] = self.df_model_name_to_hiperparams.to_dict()

    if self._top_k_best_model_idx is not None:
      self._status['BEST'] = self._status_per_grid_idx[self._top_k_best_model_idx[0]]

    return self._status

  @property
  def metadata(self):
    if self._dct_data_factories['train'] is not None:
      train_data_factory = self._dct_data_factories['train']
      if self._metadata.get('CLASSES', None) is None:
        self._metadata['CLASSES'] = train_data_factory.dataset_info['class_to_id']
      # endif

      preprocess_definitions = train_data_factory.preprocess_definitions
      self._metadata['INFERENCE_PREPROCESS_DEFINITIONS'] = [
        [preprocess_to_str(x[0]), x[1]]
        for x in preprocess_definitions
      ]
    #endif

    return self._metadata

  @property
  def grid_has_finished(self):
    grid_has_finished = False
    if self._status['NR_ALL_GRID_ITER'] is not None:
      grid_has_finished = (self._status['GRID_ITER'] == self._status['NR_ALL_GRID_ITER'])
    if self.cfg_export_model:
      grid_has_finished = grid_has_finished and self.export_done
    return grid_has_finished or self._grid_loop_exception

  def _find_data_directories(self, path_to_dataset):
    def _recursive_search_dirs(path_to_dataset, base_path=''):
      dirs = list(filter(
        lambda x: os.path.isdir(os.path.join(path_to_dataset, x)),
        os.listdir(path_to_dataset)
      ))

      subdirs = [os.path.join(base_path, d) for d in dirs]
      if len(dirs) > 0:
        for d in dirs:
          subdirs += _recursive_search_dirs(os.path.join(path_to_dataset, d), os.path.join(base_path, d))

      return subdirs

    dirs = _recursive_search_dirs(path_to_dataset)

    for k in self._dct_data_dirs.keys():
      for d in dirs:
        if k in d.lower():
          self._dct_data_dirs[k] = os.path.join(path_to_dataset, d)
          # now we break early because we found the first apparition of the 'k' dataset
          break
        # endif
      # endfor
    # endfor

    for k, v in self._dct_data_dirs.items():
      if len(v) > 0:
        self.P("'{}' dataset is found at '{}'".format(k, v), color='g')
      else:
        self.P("'{}' dataset is not available".format(k), color='r')

    return

  def _create_data_factories(self, image_height=None, image_width=None, **kwargs):
    class_def = self._data_factory_ref[2]

    for data_subset, path in self._dct_data_dirs.items():
      if len(path) == 0:
        continue

      obj = class_def(
        log=self.log,
        load_observations=self.cfg_preload_data,
        batch_size=self.cfg_batch_size,
        path_to_dataset=path,
        data_subset_name=data_subset,
        image_height=image_height,
        image_width=image_width,
        load_device=self.cfg_device_load_data,
        training_device=self.cfg_device_training,
        num_workers=self.cfg_num_workers,
        **kwargs,
      )
      obj.create()
      self._dct_data_factories[data_subset] = obj
    #endfor
    return

  def _release_data_factories(self):
    for factory in self._dct_data_factories.values():
      if factory is None:
        continue

      factory.release_device()
    #endfor
    return

  def _update_data_factories(self, **kwargs):
    for factory in self._dct_data_factories.values():
      if factory is None:
        continue

      factory.update_data_factory(**kwargs)
    # endfor


  def _maybe_test(self):
    res = None
    if self._dct_data_factories['test'] is not None:
      res = self._trainer_callbacks_obj.test_callback(
        data_generator=self._dct_data_factories['test'].data_loader,
        dataset_info=self._dct_data_factories['test'].dataset_info,
      )
    #endif

    return res

  def _maybe_dev(self):
    res = None
    if self._dct_data_factories['dev'] is not None:
      res = self._trainer_callbacks_obj.dev_callback(
        data_generator=self._dct_data_factories['dev'].data_loader,
        dataset_info=self._dct_data_factories['dev'].dataset_info,
        key='dev',
        epoch=0
      )

    # endif

    return res

  def _collect_dev_results(self, dct_grid_option, model_name, res=None):
    if res is None:
      res = self._trainer.epochs_data.values()
    else:
      columns = res.keys()
      res = res.values()
    model_results = pd.DataFrame([res])
    model_results.columns = columns
    row = model_results[model_results[self.score_key] == self._trainer.score_eval_func(model_results[self.score_key])]
    # subfolder_path = 'training_' + self.log.session_id + '_{}'.format(self.cfg_model_name)

    self.log.save_dataframe(
      df=model_results,
      fn=f'{model_name}',
      folder='output',
      subfolder_path=self.experiment_subfolder
    )

    if self.df_model_name_to_hiperparams is None:
      self.df_model_name_to_hiperparams = pd.DataFrame(
        columns=['model_name', 'model_idx', *list(dct_grid_option.keys()), *row.columns]
      )
    # endif

    self.df_model_name_to_hiperparams = pd.concat([self.df_model_name_to_hiperparams,
      pd.DataFrame([{
        'model_name': f'{model_name}',
        **dct_grid_option,
        **row.to_dict('records')[0]
      }])],
      ignore_index=True
    )

    self.df_model_name_to_hiperparams.sort_values(by=self.score_key, ascending=(self.score_mode == 'min'), inplace=True)
    self._top_k_best_model_idx = self.df_model_name_to_hiperparams['model_idx'].values[:self.cfg_keep_top_k_iterations]
    self.log.save_dataframe(
      df=self.df_model_name_to_hiperparams,
      fn='00_grid_{}_{}'.format(self._dataset_name, self.cfg_model_name),
      folder='output',
      subfolder_path=self.experiment_subfolder
    )

    self.P("Results so far:\n{}".format(self.df_model_name_to_hiperparams))
    return

  def _collect_results(self, dct_grid_option, i):
    model_results = pd.DataFrame(self._trainer.epochs_data.values())
    row = model_results[model_results[self.score_key] == self._trainer.score_eval_func(model_results[self.score_key])]
    # subfolder_path = 'training_' + self.log.session_id + '_{}'.format(self.cfg_model_name)

    self.log.save_dataframe(
      df=model_results,
      fn=f'{self.cfg_model_name}_{i}',
      folder='output',
      subfolder_path=self.experiment_subfolder
    )

    if self.df_model_name_to_hiperparams is None:
      self.df_model_name_to_hiperparams = pd.DataFrame(
        columns=['model_name', 'model_idx', *list(dct_grid_option.keys()), *row.columns]
      )
    # endif

    self.df_model_name_to_hiperparams = pd.concat([self.df_model_name_to_hiperparams,
      pd.DataFrame([{
        'model_name': f'{self.cfg_model_name}_{i}',
        'model_idx': i,
        **dct_grid_option,
        **row.to_dict('records')[0]
      }])],
      ignore_index=True
    )

    self.df_model_name_to_hiperparams.sort_values(by=self.score_key, ascending=(self.score_mode == 'min'), inplace=True)
    self._top_k_best_model_idx = self.df_model_name_to_hiperparams['model_idx'].values[:self.cfg_keep_top_k_iterations]
    self.log.save_dataframe(
      df=self.df_model_name_to_hiperparams,
      fn='00_grid_{}_{}'.format(self._dataset_name, self.cfg_model_name),
      folder='output',
      subfolder_path=self.experiment_subfolder
    )

    self.P("Results so far:\n{}".format(self.df_model_name_to_hiperparams))
    return

  def _fit(self, model, model_name, model_loss, trainer_kwargs, callbacks_kwargs):

    self._trainer = ModelTrainer(
      log=self.log,
      model=model,
      model_name=model_name,
      losses=model_loss,
      batch_size=self.cfg_batch_size,
      device=self.cfg_device_training,
      score_key=self.score_key,
      score_mode=self.score_mode,
      base_folder=self.experiment_models_folder,
      **trainer_kwargs
    )

    self._trainer_callbacks_obj = self.training_callbacks_class_def(
      log=self.log,
      owner=self._trainer,
      preprocess_before_fw_callback=self._dct_data_factories['train'].preprocess_before_forward,
      training_device=self.cfg_device_training,
      **callbacks_kwargs
    )

    self._trainer.train_on_batch = self._trainer_callbacks_obj.train_on_batch_callback

    self._trainer.evaluate = partial(
      self._trainer_callbacks_obj.evaluate_callback,
      data_generator=self._dct_data_factories['dev'].data_loader,
      dataset_info=self._dct_data_factories['dev'].dataset_info,
      key='dev'
    )

    self._trainer.evaluate_train = partial(
      self._trainer_callbacks_obj.evaluate_callback,
      data_generator=self._dct_data_factories['train'].data_loader,
      dataset_info=self._dct_data_factories['train'].dataset_info,
      key='train'
    )

    self._trainer.fit(
      th_dl=self._dct_data_factories['train'].data_loader,
      epochs=self.cfg_epochs,
      batch_index=self.cfg_batch_index
    )

    return

  def _disk_cleanup(self):
    for idx, status in self._status_per_grid_idx.items():
      if idx not in self._top_k_best_model_idx:
        try:
          os.remove(status['best_file'])
          self.P("Model files saved at iteration {} were deleted because they are not in top k best anymore.".format(idx), color='y')
        except FileNotFoundError:
          pass

        try:
          os.remove(status['best_file'] + '.optim')
        except FileNotFoundError:
          pass

    return

  def _save_model_def(self, model_kwargs, model_name):
    if not os.path.exists(self.experiment_models_folder):
      os.makedirs(self.experiment_models_folder)
    model_def_fn = "{}_def.json".format(model_name)
    self.log.save_json(
      model_kwargs,
      os.path.join(self.experiment_models_folder, model_def_fn)
    )
    return

  def get_pretrained_weights_path(self, pretrained_weights_url):
    if os.path.isfile(pretrained_weights_url):
      pretrained_weights_path = pretrained_weights_url
    else:
      # TODO: review this in order to be able to use minio
      try:
        saved_files, _ = self.log.maybe_download(
          url=pretrained_weights_url,
          fn=f'{self.cfg_model_name}_pretrained_weights.th',
          target='models',
          unzip=False
        )
        pretrained_weights_path = saved_files[0]
      except Exception as e:
        self.log.P(f'Could not download pretrained weights from {pretrained_weights_url}', color='r')
        pretrained_weights_path = None
      # endtry download
    # endif pretrained_weights_url is a local file
    return pretrained_weights_path

  def trace_preprocess_method(self, inputs, device, preprocess_method, **kwargs):
    inputs = [inp.to(device) for inp in inputs]
    if preprocess_method is not None:
      inputs = [preprocess_method(inp, **kwargs) for inp in inputs]
    th_inputs = th.stack(inputs)
    return th_inputs


  def trace_best(self):
    """
    Method that will trace the best model.
    Returns
    -------
    path : str, path to the best model if trace was successful, None otherwise
    """
    res_path = None
    best_weights_path = self.status['BEST']['best_file']
    best_model = self.model_factory_class_def(**self.status['BEST']['model_kwargs']).to(self.cfg_device_training)
    best_model.load_state_dict(th.load(best_weights_path, map_location=self.cfg_device_training))
    best_model.eval()

    n_samples = 5
    observations = []
    labels = []
    indexes = []
    data_factory = None
    for set_key in ['train', 'dev', 'test']:
      if self._dct_data_factories[set_key] is not None:
        curr_obs = self._dct_data_factories[set_key].data_loader.dataset._observations
        if len(curr_obs) > n_samples:
          data_factory = self._dct_data_factories[set_key]
          observations = curr_obs[:n_samples]
          labels = data_factory.data_loader.dataset._labels[:n_samples]
          break
        # endif enough samples
      # endif existing set
    # endfor set_key
    if len(observations) == 0:
      self.P("No observations found in the dataset. Cannot trace the model.")
      return res_path
    # endif no observations
    if isinstance(observations[0], str):
      # If the observations are paths to inputs, we need to load them
      observations = [data_factory.load_x_and_y_callback(observations, labels, i)[0] for i in range(n_samples)]
    # endif not loaded observations
    preprocess_method = self._dct_data_factories[set_key]._before_forward_preprocess
    if observations[0].device != self.cfg_device_training:
      observations = [obs.to(self.cfg_device_training) for obs in observations]
    traced_model_name = '__'.join([self.log.now_str()[:-6], self.cfg_model_name])
    input_shape = tuple(observations[0].permute(1, 2, 0).shape)
    tracer = BaseTorchScripter(
      log=self.log, model=best_model, model_name=traced_model_name,
      preprocess_method=self.trace_preprocess_method, input_shape=input_shape
    )
    res_path = tracer.generate(
      inputs=observations, batch_size=1, device=self.cfg_device_training,
      preprocess_method=preprocess_method, to_test=True, test_batch_size=n_samples
    )
    return res_path

  def run(self, start_iter=0, end_iter=None):
    success = False
    self.maybe_disk_preprocess_dataset()
    self._time_start_run = time()
    self._grid_loop_exception = False
    dct_grid_search = self.get_grid_search() or self.default_grid_search
    lst_data_params = dct_grid_search.get('DATA_PARAMS', [])
    lst_callbacks_params = dct_grid_search.get('CALLBACKS_PARAMS', [])
    lst_trainer_params = dct_grid_search.get('TRAINER_PARAMS', [])
    lst_model_params = dct_grid_search.get('MODEL_PARAMS', None)
    lst_exceptions = dct_grid_search.get('EXCEPTIONS', None)
    lst_fixed = dct_grid_search.get('FIXED', None)
    lst_keep_non_model_params = dct_grid_search.get('KEEP_NON_MODEL_PARAMS', [])
    all_grid_options = self.log.get_grid_iterations(
      params_grid=dct_grid_search['GRID'],
      priority_keys=lst_data_params,
      exceptions=lst_exceptions,
      fixed=lst_fixed,
      verbose=1
    )

    if self.cfg_end_grid_idx is not None:
      all_grid_options = all_grid_options[:self.cfg_end_grid_idx]
    if self.cfg_start_grid_idx is not None:
      all_grid_options = all_grid_options[self.cfg_start_grid_idx:]

    self._status['NR_ALL_GRID_ITER'] = len(all_grid_options)
    self.P("Grid search for model base name: {}".format(self.cfg_model_name), color='b')

    lst_non_model_params = lst_callbacks_params + lst_trainer_params + lst_data_params
    if lst_model_params is None:
      lst_model_params = list((set(list(dct_grid_search['GRID'].keys())) - set(lst_non_model_params)).union(set(lst_keep_non_model_params)))

    iter_times = []
    iteration = 0
    last_data_params = None
    if end_iter is None:
      end_iter = len(all_grid_options)
    try:
      grid_finished = False
      for i, dct_grid_option in enumerate(all_grid_options):
        if i < start_iter:
          continue
        if i >= end_iter:
          continue
        iteration = i+1
        time_start_iter = time()
        self.log.reset_seeds(seed=42)
        model_name = self.cfg_model_name + '_{}'.format(i)
        lst_grid_dict = ["{}:{}".format(k, v) for k, v in dct_grid_option.items()]
        str_grid_dict = "  ".join(lst_grid_dict)
        self.P("Training grid option {}/{}: {}".format(iteration, len(all_grid_options), str_grid_dict), color='g')

        data_kwargs = {k: dct_grid_option[k] for k in lst_data_params}
        if last_data_params != data_kwargs:
          if last_data_params is None:
            self._create_data_factories(**data_kwargs)
          else:
            try:
              # Try to update data factories
              self._update_data_factories(**data_kwargs)
            except NotImplementedError:
              # If update method not implemented -> reload them
              self._release_data_factories()
              self._create_data_factories(**data_kwargs)
            #end try/except
          #endif
          last_data_params = data_kwargs
        #endif

        trainer_kwargs = {k : dct_grid_option[k] for k in lst_trainer_params}
        callbacks_kwargs = {k : dct_grid_option[k] for k in lst_callbacks_params}
        model_kwargs = {k: dct_grid_option[k] for k in lst_model_params}

        self._save_model_def(model_kwargs, model_name)
        if inspect.getfullargspec(self.model_factory_class_def.__init__).varkw is not None:
          model_kwargs = dct_grid_option

        model = self.model_factory_class_def(**model_kwargs).to(self.cfg_device_training)
        model_loss = self.model_loss(**dct_grid_option)
        skip_iteration = False
        pretrained_weights_url = trainer_kwargs.get('pretrained_weights', None)
        pretrained_weights_url = self.cfg_pretrained_weights_url if pretrained_weights_url is None else pretrained_weights_url
        if pretrained_weights_url is not None:
          pretrained_weights_path = self.get_pretrained_weights_path(pretrained_weights_url)
          if pretrained_weights_path is not None:
            try:
              model.load_state_dict(th.load(pretrained_weights_path, map_location=self.cfg_device_training))
              self.log.P(f'Successfully loaded pretrained weights from {pretrained_weights_path}', color='g')
            except Exception as e:
              skip_iteration = True
              self.log.P(
                f'Could not load pretrained weights from {pretrained_weights_path}. Skipping current iteration',
                color='r')
            # endtry load pretrained weights
          # endif pretrained weights successfully located
        # endif pretrained_weights_url

        if not skip_iteration:
          self._fit(
            model=model,
            model_name=model_name,
            model_loss=model_loss,
            trainer_kwargs=trainer_kwargs,
            callbacks_kwargs=callbacks_kwargs
          )

          self._status_per_grid_idx[i] = self._trainer.training_status.copy()
          self._status_per_grid_idx[i]['model_kwargs'] = model_kwargs
          self._status_per_grid_idx[i]['dct_grid_option'] = dct_grid_option

          dct_test_results = self._maybe_test()
          if dct_test_results is not None:
            self._trainer.epochs_data[self._trainer.training_status['best_epoch']] = {
              **self._trainer.epochs_data[self._trainer.training_status['best_epoch']],
              **dct_test_results
            }
          self._collect_results(dct_grid_option, i)
          iter_times.append(time()-time_start_iter)
          self._status['GRID_ITER'] = iteration
          self._status['REMAINING'] = np.mean(iter_times) * (self._status['NR_ALL_GRID_ITER'] - self._status['GRID_ITER'])
        # endif skip_iteration
        self._disk_cleanup()
        model.cpu()
        del model
      #endfor
      grid_finished = True
      if self.cfg_export_model:
        trace_path = self.trace_best()
        if trace_path is not None:
          self._metadata['MODEL_EXPORT'] = trace_path
        self.export_done = True
      success = True
    except ct.ForceStopException as e:
      if not grid_finished or not self.cfg_export_model:
        self.P(f"Grid search stopped at iteration {iteration}/{self._status['NR_ALL_GRID_ITER']}: {e}")
      else:
        self.P(
          f"Force stop during model exporting after grid search finished at iteration "
          f"{iteration}/{self._status['NR_ALL_GRID_ITER']}: {e}"
        )
      # endif grid_finished
    except Exception as e:
      if not grid_finished or not self.cfg_export_model:
        msg = (
          f"Grid Exception at iteration {iteration}/{self._status['NR_ALL_GRID_ITER']}:\n{traceback.format_exc()}\n"
          f"{self.log.get_error_info(return_err_val=True)}"
        )
      else:
        msg = (
          f"Grid Exception during model exporting after grid search finished at iteration "
          f"{iteration}/{self._status['NR_ALL_GRID_ITER']}:\n{traceback.format_exc()}\n"
          f"{self.log.get_error_info(return_err_val=True)}"
        )
      # endif grid_finished
      self.P(msg, color='r')
      self._grid_loop_exception = True
    # end try-except

    self._release_data_factories()

    return success



"""
pipeline start example:
```json
{
  "EE_ID" : "aid_mob",
  "ACTION" : "UPDATE_CONFIG",
  "PAYLOAD" : {
      "NAME" : "llm_test",
      "TYPE" : "Void",
      "PLUGINS" : [
          {
              "SIGNATURE" : "LLM_CODEGEN_01",
              "INSTANCES" : [
                  {
                      "INSTANCE_ID" : "LLM_CODEGEN_01_def1"
                  }
              ]
          }
      ]
  },
  "INITIATOR_ID" : "Explorer_xxxx",
  "SESSION_ID" : "Explorer_xxxx20230918144709767060",
  "EE_SIGN" : "MEUCIBESpz-w3tdV01nuJ9mbxm5THwUeqnKQ-knZ4QZfnutfAiEAweX_VHpbkHUgrC7W7484f4rVnLSLmEyHdcWXC5iISGY=",
  "EE_SENDER" : "0xai_Au7FmjsXxtgLTFPV1ux9bI5a7UhAIK6azRfS0FGTMf2T",
  "EE_HASH" : "52189177d299b393f8ae04171f646a81cb424375230e3b0dc3bf2a4022550fa1"
}	
```
 
request example:
```json
	{
		"EE_ID" : "target-node",
		"ACTION" : "UPDATE_PIPELINE_INSTANCE",
		"PAYLOAD" : {
			"NAME" : "LLM_PIPELINE",
			"SIGNATURE" : "LLM_CODEGEN_01",
			"INSTANCE_ID" : "DEFAULT",
			"INSTANCE_CONFIG" : {
				"INSTANCE_COMMAND": {
					"history": [
						{
							"request" : "write the php script for connecting to a mqtt server",
							"response" : "<?php \n require("path/to/phpMQTT.php");..."
						},
						{
							"request" : "add a callback that filters all the messages that ..."",
							"response" : "function filterMessages($topic, $msg) {..."
						}
					],
					"request" : "add a hello world example",
					"request_id" : "1234"						
				}
			}
		},
		"INITIATOR_ID" : "Explorer_xxxx",
		"SESSION_ID" : "Explorer_xxxx20230918140745309935",
		"EE_SIGN" : "MEUCIHGvyPrdu5OmrqfjOsmZi5fpI7x0vQunRFbgpJwWTbqyAiEAvrZOhMnxXrbVchmgdmZ6q9xoQMVa6wS30R_qIRHnuOk=",
		"EE_SENDER" : "0xai_Au7FmjsXxtgLTFPV1ux9bI5a7UhAIK6azRfS0FGTMf2T",
		"EE_HASH" : "2aa7ada77792e421bcc8ecd31c6f270510f1dd8c6371d1f16cc5e0ba3474d70a"
	}
```

  
"""


from naeural_core.business.base import BasePluginExecutor as BasePlugin


__VER__ = '0.1.0.0'

_CONFIG = {

  # mandatory area
  **BasePlugin.CONFIG,

  # our overwritten props
  # 'AI_ENGINE'     : "llm_server",
  'OBJECT_TYPE'   : [],
  'PROCESS_DELAY' : 10,
  

  'VALIDATION_RULES' : {
    **BasePlugin.CONFIG['VALIDATION_RULES'],
  },  
}

class LlmCodegen01Plugin(BasePlugin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    super(LlmCodegen01Plugin, self).__init__(**kwargs)
    return
  
  
  
  def __process_request(self, query, request_id, history, user):
    static_result = f"""
    #
    # Processing node {self.eeid}:{self.node_addr}
    #
    # This is a static code example generated  for testing purposes
    # based on request_id {request_id} from user {user}:
    \"\"\"
    {query}
    \"\"\"
    
    if __name__ == '__main__':
			print("Hello World!")
    """
    if user is not None:
      # process with persistance
      result = self.chatapi_ask(
        question=query,
        persona="codegen",
        user=user,
        set_witness=False,
        personas_folder='./core/utils/openai/personas'
      )
    else:
      result = static_result
    return result
  
  def on_command(self, data, **kwargs):
    request = data.get('request', None)
    request_id = data.get('request_id', None)
    history = data.get('history', [])
    user_id = data.get('user', None)
    temperature = data.get('temperature', None)

    resp = self.__process_request(
      query=request, 
      request_id=request_id,
      history=history,
      user=user_id,
    )
    
    self.P("Generated code:\n{}".format(resp))
    
    self.create_and_send_payload(
      codegen_response=resp,
      codegen_request_id=request_id,
      codegen_history=history,
      codegen_user=user_id,
      codegen_temperature=temperature,
		)
    return

  def _process(self):
    payload = None
    return payload


from naeural_core.utils.img_utils import maybe_prepare_img_payload
from naeural_core.heavy_ops.base import BaseHeavyOp

class ImageCompressionHeavyOp(BaseHeavyOp):

  def __init__(self, **kwargs):
    super(ImageCompressionHeavyOp, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    # should run only in comm thread and not async as it replaces inplace the
    # 'IMG' with the base64 string
    assert not self.comm_async 

  def _register_payload_operation(self, payload):
    """
    this plugin must modify the payload before sending to outside
    """
    return payload # we just return original dict so we can change it inplace

  def _process_dct_operation(self, dct):
    maybe_prepare_img_payload(sender=self, dct=dct)
    return



from ratio1 import BaseDecentrAIObject
from naeural_core.core_logging import Logger
from collections import deque
import traceback

class DecentrAIObject(BaseDecentrAIObject):
  """
  Generic class

  Instructions:

    1. use `super().__init__(**kwargs)` at the end of child `__init__`
    2. define `startup(self)` method for the child class and call 
       `super().startup()` at beginning of `startup()` method

      OR

    use `super().__init__(**kwargs)` at beginning of child `__init__` and then
    you can safely proceed with other initilization 

  """

  def __init__(self, log: Logger,
               DEBUG=False,
               show_prefixes=False,
               prefix_log=None,
               maxlen_notifications=None,
               log_at_startup=False,
               **kwargs):
    self._messages = deque(maxlen=maxlen_notifications)
    super(DecentrAIObject, self).__init__(
        log=log,
        DEBUG=DEBUG,
        show_prefixes=show_prefixes,
        prefix_log=prefix_log,
        maxlen_notifications=maxlen_notifications,
        log_at_startup=log_at_startup,
        **kwargs
    )
    return

  def _parse_config_data(self, *args, **kwargs):
    """
    args: keys that are used to prune the config_data. Examples:
                1. args=['TEST'] -> kwargs will be searched in
                                    log.config_data['TEST']
                2. args=['TEST', 'K1'] -> kwargs will be searched in
                                          log.config_data['TEST']['K1']
    kwargs: dictionary of k:v pairs where k is a parameter and v is its value.
            If v is None, then k will be searched in logger config data in order to set
            the value specified in json.
            Finally, the method will set the final value to a class attribute named 
            exactly like the key.
    """
    cfg = self.log.config_data
    for x in args:
      if x is not None:
        cfg = cfg[x]

    for k, v in kwargs.items():
      if v is None and k in cfg:
        v = cfg[k]

      setattr(self, k, v)

    return

  def _create_notification(
    self,
    notif,
    msg,
    info=None,
    stream_name=None,
    autocomplete_info=False,
    notif_code=None,
    notif_tag=None,
    ct=None,
    **kwargs
  ):
    body = {
      'MODULE': self.__class__.__name__
    }

    if hasattr(self, '__version__'):
      body['VERSION'] = self.__version__

    if autocomplete_info and info is None:
      err_info = self.log.get_error_info(return_err_val=True)
      trace_info = traceback.format_exc()
      if err_info[0] != '':
        info = "* Log error info:\n{}\n* Traceback:\n{}".format(
          err_info,
          trace_info,
        )
    # endif

    error_code = None
    if notif_code is not None and notif_code < 0:
      error_code = notif_code

    if notif_code is not None and notif_tag is None and ct is not None:
      notif_tag = ct.NOTIFICATION_CODES.TAGS.get(notif_code, None)
    if notif_tag is not None and notif_code is None and ct is not None:
      notif_code = ct.NOTIFICATION_CODES.CODES.get(notif_tag, None)

    body['NOTIFICATION_TYPE'] = notif
    body['NOTIFICATION'] = msg[:255]
    body['NOTIFICATION_CODE'] = notif_code
    body['NOTIFICATION_TAG'] = notif_tag
    body['INFO'] = info
    body['STREAM_NAME'] = stream_name
    body['TIMESTAMP'] = self.log.now_str(nice_print=True, short=False)
    body['ERROR_CODE'] = error_code
    body = {
      **body,
      **{k.upper(): v for k, v in kwargs.items()}
    }
    self._messages.append(body)
    return

  def get_notifications(self):
    lst = []
    while len(self._messages) > 0:
      lst.append(self._messages.popleft())
    return lst

  def get_cmd_handlers(self, update=False):
    if hasattr(self, 'COMMANDS') and isinstance(getattr(self, 'COMMANDS'), dict):
      COMMANDS = self.COMMANDS.copy()
    else:
      COMMANDS = {}
    for k in dir(self):
      if k.startswith('cmd_handler_'):
        cmd = k.replace('cmd_handler_', '').upper()
        if cmd not in COMMANDS:
          COMMANDS[cmd] = getattr(self, k)
    if update:
      self.COMMANDS = COMMANDS
    return COMMANDS

  def run_cmd(self, cmd, **kwargs):
    res = None
    cmd = cmd.upper()
    dct_cmds = self.get_cmd_handlers()
    if cmd in dct_cmds:
      func = dct_cmds[cmd]
      res = func(**kwargs)
    else:
      print("Received unk command '{}'".format(cmd))
    return res



"""
    SORT: A Simple, Online and Realtime Tracker
    Copyright (C) 2016-2020 Alex Bewley alex@bewley.ai

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function

import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from skimage import io

import glob
import time
import argparse 
from naeural_core.utils.kalman_filter import KalmanFilter

np.random.seed(0)


def linear_assignment(cost_matrix):
  try:
    import lap
    _, x, y = lap.lapjv(cost_matrix, extend_cost=True)
    return np.array([[y[i],i] for i in x if i >= 0]) #
  except ImportError:
    from scipy.optimize import linear_sum_assignment
    x, y = linear_sum_assignment(cost_matrix)
    return np.array(list(zip(x, y)))


def iou_batch(bb_test, bb_gt):
  """
  From SORT: Computes IOU between two bboxes in the form [x1,y1,x2,y2]
  """
  bb_gt = np.expand_dims(bb_gt, 0)
  bb_test = np.expand_dims(bb_test, 1)
  
  xx1 = np.maximum(bb_test[..., 0], bb_gt[..., 0])
  yy1 = np.maximum(bb_test[..., 1], bb_gt[..., 1])
  xx2 = np.minimum(bb_test[..., 2], bb_gt[..., 2])
  yy2 = np.minimum(bb_test[..., 3], bb_gt[..., 3])
  w = np.maximum(0., xx2 - xx1)
  h = np.maximum(0., yy2 - yy1)
  wh = w * h
  o = wh / ((bb_test[..., 2] - bb_test[..., 0]) * (bb_test[..., 3] - bb_test[..., 1])                                      
    + (bb_gt[..., 2] - bb_gt[..., 0]) * (bb_gt[..., 3] - bb_gt[..., 1]) - wh)                                              
  return(o)  


def convert_bbox_to_z(bbox):
  """
  Takes a bounding box in the form [x1,y1,x2,y2] and returns z in the form
    [x,y,s,r] where x,y is the centre of the box and s is the scale/area and r is
    the aspect ratio
  """
  w = bbox[2] - bbox[0]
  h = bbox[3] - bbox[1]
  x = bbox[0] + w/2.
  y = bbox[1] + h/2.
  s = w * h    #scale is just area
  r = w / float(h)
  return np.array([x, y, s, r]).reshape((4, 1))


def convert_x_to_bbox(x,score=None):
  """
  Takes a bounding box in the centre form [x,y,s,r] and returns it in the form
    [x1,y1,x2,y2] where x1,y1 is the top left and x2,y2 is the bottom right
  """
  w = np.sqrt(x[2] * x[3])
  h = x[2] / w
  if(score==None):
    return np.array([x[0]-w/2.,x[1]-h/2.,x[0]+w/2.,x[1]+h/2.]).reshape((1,4))
  else:
    return np.array([x[0]-w/2.,x[1]-h/2.,x[0]+w/2.,x[1]+h/2.,score]).reshape((1,5))


class KalmanBoxTracker(object):
  """
  This class represents the internal state of individual tracked objects observed as bbox.
  """
  count = 0
  def __init__(self,bbox):
    """
    Initialises a tracker using initial bounding box.
    """
    #define constant velocity model
    self.kf = KalmanFilter(dim_x=7, dim_z=4) 
    self.kf.F = np.array([[1,0,0,0,1,0,0],[0,1,0,0,0,1,0],[0,0,1,0,0,0,1],[0,0,0,1,0,0,0],  [0,0,0,0,1,0,0],[0,0,0,0,0,1,0],[0,0,0,0,0,0,1]])
    self.kf.H = np.array([[1,0,0,0,0,0,0],[0,1,0,0,0,0,0],[0,0,1,0,0,0,0],[0,0,0,1,0,0,0]])

    self.kf.R[2:,2:] *= 10.
    self.kf.P[4:,4:] *= 1000. #give high uncertainty to the unobservable initial velocities
    self.kf.P *= 10.
    self.kf.Q[-1,-1] *= 0.01
    self.kf.Q[4:,4:] *= 0.01

    self.kf.x[:4] = convert_bbox_to_z(bbox)
    self.time_since_update = 0
    self.id = KalmanBoxTracker.count
    KalmanBoxTracker.count += 1
    self.history = []
    self.hits = 0
    self.hit_streak = 0
    self.age = 0

  def update(self,bbox):
    """
    Updates the state vector with observed bbox.
    """
    self.time_since_update = 0
    self.history = []
    self.hits += 1
    self.hit_streak += 1
    self.kf.update(convert_bbox_to_z(bbox))

  def predict(self):
    """
    Advances the state vector and returns the predicted bounding box estimate.
    """
    if((self.kf.x[6]+self.kf.x[2])<=0):
      self.kf.x[6] *= 0.0
    self.kf.predict()
    self.age += 1
    if(self.time_since_update>0):
      self.hit_streak = 0
    self.time_since_update += 1
    self.history.append(convert_x_to_bbox(self.kf.x))
    return self.history[-1]

  def get_state(self):
    """
    Returns the current bounding box estimate.
    """
    return convert_x_to_bbox(self.kf.x)


def associate_detections_to_trackers(detections,trackers,iou_threshold = 0.3):
  """
  Assigns detections to tracked object (both represented as bounding boxes)

  Returns 3 lists of matches, unmatched_detections and unmatched_trackers
  """
  if(len(trackers)==0):
    return np.empty((0,2),dtype=int), np.arange(len(detections)), np.empty((0,5),dtype=int)

  iou_matrix = iou_batch(detections, trackers)

  if min(iou_matrix.shape) > 0:
    a = (iou_matrix > iou_threshold).astype(np.int32)
    if a.sum(1).max() == 1 and a.sum(0).max() == 1:
        matched_indices = np.stack(np.where(a), axis=1)
    else:
      matched_indices = linear_assignment(-iou_matrix)
  else:
    matched_indices = np.empty(shape=(0,2))

  unmatched_detections = []
  for d, det in enumerate(detections):
    if(d not in matched_indices[:,0]):
      unmatched_detections.append(d)
  unmatched_trackers = []
  for t, trk in enumerate(trackers):
    if(t not in matched_indices[:,1]):
      unmatched_trackers.append(t)

  #filter out matched with low IOU
  matches = []
  for m in matched_indices:
    if(iou_matrix[m[0], m[1]]<iou_threshold):
      unmatched_detections.append(m[0])
      unmatched_trackers.append(m[1])
    else:
      matches.append(m.reshape(1,2))
  if(len(matches)==0):
    matches = np.empty((0,2),dtype=int)
  else:
    matches = np.concatenate(matches,axis=0)

  return matches, np.array(unmatched_detections), np.array(unmatched_trackers)


class Sort(object):
  def __init__(self, max_age=1, min_hits=3, iou_threshold=0.3):
    """
    Sets key parameters for SORT
    """
    self.max_age = max_age
    self.min_hits = min_hits
    self.iou_threshold = iou_threshold
    self.trackers = []
    self.frame_count = 0

  def update(self, dets=np.empty((0, 5))):
    """
    Params:
      dets - a numpy array of detections in the format [[x1,y1,x2,y2,score],[x1,y1,x2,y2,score],...]
    Requires: this method must be called once for each frame even with empty detections (use np.empty((0, 5)) for frames without detections).
    Returns the a similar array, where the last column is the object ID.

    NOTE: The number of objects returned may differ from the number of detections provided.
    """
    self.frame_count += 1
    # get predicted locations from existing trackers.
    trks = np.zeros((len(self.trackers), 5))
    to_del = []
    ret = []
    for t, trk in enumerate(trks):
      pos = self.trackers[t].predict()[0]
      trk[:] = [pos[0], pos[1], pos[2], pos[3], 0]
      if np.any(np.isnan(pos)):
        to_del.append(t)
    trks = np.ma.compress_rows(np.ma.masked_invalid(trks))
    for t in reversed(to_del):
      self.trackers.pop(t)
    matched, unmatched_dets, unmatched_trks = associate_detections_to_trackers(dets,trks, self.iou_threshold)

    # update matched trackers with assigned detections
    for m in matched:
      self.trackers[m[1]].update(dets[m[0], :])

    # create and initialise new trackers for unmatched detections
    for i in unmatched_dets:
        trk = KalmanBoxTracker(dets[i,:])
        self.trackers.append(trk)
    i = len(self.trackers)
    for trk in reversed(self.trackers):
        d = trk.get_state()[0]
        if (trk.time_since_update < 1) and (trk.hit_streak >= self.min_hits or self.frame_count <= self.min_hits):
          ret.append(np.concatenate((d,[trk.id+1])).reshape(1,-1)) # +1 as MOT benchmark requires positive
        i -= 1
        # remove dead tracklet
        if(trk.time_since_update > self.max_age):
          self.trackers.pop(i)
    if(len(ret)>0):
      return np.concatenate(ret)
    return np.empty((0,5))

def parse_args():
    """Parse input arguments."""
    parser = argparse.ArgumentParser(description='SORT demo')
    parser.add_argument('--display', dest='display', help='Display online tracker output (slow) [False]',action='store_true')
    parser.add_argument("--seq_path", help="Path to detections.", type=str, default='data')
    parser.add_argument("--phase", help="Subdirectory in seq_path.", type=str, default='train')
    parser.add_argument("--max_age", 
                        help="Maximum number of frames to keep alive a track without associated detections.", 
                        type=int, default=1)
    parser.add_argument("--min_hits", 
                        help="Minimum number of associated detections before track is initialised.", 
                        type=int, default=3)
    parser.add_argument("--iou_threshold", help="Minimum IOU for match.", type=float, default=0.3)
    args = parser.parse_args()
    return args

if __name__ == '__main__':
  # all train
  args = parse_args()
  display = args.display
  phase = args.phase
  total_time = 0.0
  total_frames = 0
  colours = np.random.rand(32, 3) #used only for display
  if(display):
    if not os.path.exists('mot_benchmark'):
      print('\n\tERROR: mot_benchmark link not found!\n\n    Create a symbolic link to the MOT benchmark\n    (https://motchallenge.net/data/2D_MOT_2015/#download). E.g.:\n\n    $ ln -s /path/to/MOT2015_challenge/2DMOT2015 mot_benchmark\n\n')
      exit()
    plt.ion()
    fig = plt.figure()
    ax1 = fig.add_subplot(111, aspect='equal')

  if not os.path.exists('output'):
    os.makedirs('output')
  pattern = os.path.join(args.seq_path, phase, '*', 'det', 'det.txt')
  for seq_dets_fn in glob.glob(pattern):
    mot_tracker = Sort(max_age=args.max_age, 
                       min_hits=args.min_hits,
                       iou_threshold=args.iou_threshold) #create instance of the SORT tracker
    seq_dets = np.loadtxt(seq_dets_fn, delimiter=',')
    seq = seq_dets_fn[pattern.find('*'):].split(os.path.sep)[0]
    
    with open(os.path.join('output', '%s.txt'%(seq)),'w') as out_file:
      print("Processing %s."%(seq))
      for frame in range(int(seq_dets[:,0].max())):
        frame += 1 #detection and frame numbers begin at 1
        dets = seq_dets[seq_dets[:, 0]==frame, 2:7]
        dets[:, 2:4] += dets[:, 0:2] #convert to [x1,y1,w,h] to [x1,y1,x2,y2]
        total_frames += 1

        if(display):
          fn = os.path.join('mot_benchmark', phase, seq, 'img1', '%06d.jpg'%(frame))
          im =io.imread(fn)
          ax1.imshow(im)
          plt.title(seq + ' Tracked Targets')

        start_time = time.time()
        trackers = mot_tracker.update(dets)
        cycle_time = time.time() - start_time
        total_time += cycle_time

        for d in trackers:
          print('%d,%d,%.2f,%.2f,%.2f,%.2f,1,-1,-1,-1'%(frame,d[4],d[0],d[1],d[2]-d[0],d[3]-d[1]),file=out_file)
          if(display):
            d = d.astype(np.int32)
            ax1.add_patch(patches.Rectangle((d[0],d[1]),d[2]-d[0],d[3]-d[1],fill=False,lw=3,ec=colours[d[4]%32,:]))

        if(display):
          fig.canvas.flush_events()
          plt.draw()
          ax1.cla()

  print("Total Tracking took: %.3f seconds for %d frames or %.1f FPS" % (total_time, total_frames, total_frames / total_time))

  if(display):
    print("Note: to get real runtime results run without the option: --display")



from naeural_core.comm import MQTTWrapper
from naeural_core.core_logging import SBLogger
from collections import deque
from time import sleep, time

if __name__ == '__main__':

  log = SBLogger()
  url_config_app = 'https://www.dropbox.com/s/bcoawurb1tyq7ma/config_app_paho_staging.txt?dl=1'

  saved_files, _ = log.maybe_download(
    url=url_config_app,
    fn='tmp/config_app_paho.txt',
    force_download=True,
    target='output'
  )

  config_app = log.load_json(saved_files[0])
  config_communication = config_app['COMMUNICATION']

  log.P("Server params:", color='b')
  for k, v in config_communication['PARAMS'].items():
    log.P(" * {}: {}".format(k, v), color='b')

  config_communication['PARAMS']['EE_ID'] = 'icsulescu'

  recv_buffer_1 = deque(maxlen=1000)
  receiver1 = MQTTWrapper(
    log=log, config=config_communication['PARAMS'], recv_buff=recv_buffer_1, recv_channel_name='NOTIF_CHANNEL')

  _dct_conn_receiver1 = receiver1.server_connect()
  log.P("Recv1 conn: {}".format(_dct_conn_receiver1))

  _dct_subscribe_receiver1 = receiver1.subscribe()
  log.P("Recv1 subs: {}".format(_dct_subscribe_receiver1))

  all_notifications = []

  def read_all(receiver, receiver_name):
    log.P("Flushing the queue for receiver {} ...".format(receiver_name))
    prev_len = len(receiver._recv_buff)
    while True:
      receiver.receive()
      crt_len = len(receiver._recv_buff)
      if crt_len == prev_len:
        break

      prev_len = crt_len
      sleep(0.1)
    # endwhile

    log.P("  Flush done, there were {} messages in the queue:".format(crt_len))
    while len(receiver._recv_buff) > 0:
      x = receiver._recv_buff.popleft()
      all_notifications.append(x)
      log.P("  {}".format(str(x)[:80]))
  # enddef

  start_loop = time()
  while True:
    read_all(receiver1, '1')
    sleep(5)

    if time() - start_loop >= 3 * 60:
      break
  # endwhile




from naeural_core import DecentrAIObject
import abc
import time as tm
import pandas as pd



class BaseConnector(DecentrAIObject):
  def __init__(self, log, config, verbose=True, **kwargs):
    self.readers = {}
    self.config = config
    self.verbose = verbose
    super(BaseConnector, self).__init__(log=log, prefix_log='[CONN]', **kwargs)
    return

  def connect(self, nr_retries=None):
    if nr_retries is None:
      nr_retries = 5

    self.P("Connecting ...")
    count = 0
    while count < nr_retries:
      count += 1
      try:
        self._connect(**self.config['CONNECT_PARAMS'])
        self.P("connection created.", color='g')
        break
      except Exception as err:
        self.P("ERROR! connection failed\n{}".format(err))
        tm.sleep(0.5)
      #end try-except
    #endwhile
    return

  def maybe_create_reader(self, reader_name, **kwargs):
    if len(kwargs) == 0:
      query_params = self.config['QUERY_PARAMS'].get(reader_name, None)
    else:
      query_params = kwargs
      
    if query_params is None:
      self.P("Could not create reader {}. No query params provided or found in config.".format(reader_name), color='r')

    reader = self.readers.get(reader_name, None)
    if reader is None:
      try:
        reader = self._create_reader(**query_params)
        self.readers[reader_name] = reader
        if self.verbose:
          self.P("Successfully created reader {}".format(reader_name), color='g')
      except Exception as e:
        self.P("Error creating reader {}\n{}".format(reader_name, e), color='r')
    #endif
    return

  
  def get_data(self, **kwargs):
    kwargs['chunksize'] = None
    temp_name = self.log.now_str()
    self.maybe_create_reader(temp_name, **kwargs)
    data = self.readers[temp_name].copy()
    self.delete_reader(temp_name)
    return data

  def create_readers_from_config(self):
    reader_names = list(self.config['QUERY_PARAMS'].keys())
    for reader_name in reader_names:
      self.create_reader(reader_name)
    return

  def get_data_generator_from_reader(self, reader_name):
    self.maybe_create_reader(reader_name)
    reader = self.readers.get(reader_name, None)
    if reader is None:
      return
    if isinstance(reader, pd.DataFrame):
      raise ValueError("Attempted to use whole dataframe {} as generator failed. Please use `chunksize` to create generators".format(
        reader.shape))
    self.P("Iterating chunks for reader {} ...".format(reader_name), color='b')
    for i, df_chunk in enumerate(reader):
      d1, d2 = df_chunk.shape
      print_msg = "Chunk events: {}".format(d1)
      if i == 0:
        print_msg += " / cols: {}".format(d2)

      self.P(print_msg, color='b')

      yield df_chunk

  def get_all_data_generators(self):
    reader_names = list(self.config['QUERY_PARAMS'].keys())
    for reader_name in reader_names:
      for df_chunk in self.get_data_generator_from_reader(reader_name):
        yield reader_name, df_chunk

  def get_data_from_reader(self, reader_name):
    lst_df_chunks = []
    for df_chunk in self.get_data_generator_from_reader(reader_name):
      lst_df_chunks.append(df_chunk)
    if len(lst_df_chunks) > 1:
      return pd.concat(lst_df_chunks)
    else:
      if len(lst_df_chunks) == 1:
        return lst_df_chunks[0]
      else:
        return None

  def get_all_readers_data(self):
    reader_names = list(self.config['QUERY_PARAMS'].keys())
    dct_all_data = {}
    for reader_name in reader_names:
      dct_all_data[reader_name] = self.get_data_from_reader(reader_name)
    return dct_all_data

  def _connect(self, **kwargs):
    return
  
  def delete_reader(self, reader_name):
    if reader_name in self.readers:
      del self.readers[reader_name]
    return

  @abc.abstractmethod
  def _create_reader(self, **kwargs):
    raise NotImplementedError


__version__ = '1.0.0.0'

import os
import azureml
from ratio1 import BaseDecentrAIObject

from azureml.core import Workspace
from azureml.core.model import Model, InferenceConfig
from azureml.core.environment import Environment
from azureml.core.webservice import AciWebservice, Webservice
from azureml.core.authentication import ServicePrincipalAuthentication
try:
  from .enums import AzureLocationEnum
except:
  from azureml.enums import AzureLocationEnum
from time import time

class AuthenticationService(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def service_principal_authentication(self, tenant_id, service_principal_id, service_principal_password):
    """
    This method allows authentication into AzureML with service principal method

    Parameters
    ----------
    tenant_id : string
      AzureML tenant id .
    service_principal_id : TYPE
      AzureML client id (email).
    service_principal_password : TYPE
      AzureML client password.
    -------
    ServicePrincipalAuthentication object or throws exception

    """
    self.log.p('Making service principal authentication...')
    sp = ServicePrincipalAuthentication(
      tenant_id=tenant_id,
      service_principal_id=service_principal_id,
      service_principal_password=service_principal_password)
    sp.get_authentication_header()
    self.log.p('Done authentication', show_time=True)
    return sp


class WorkspaceService(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def create(self, auth, subscription_id, resource_group_name, workspace_name, location=AzureLocationEnum.EuropeWest.value):
    """
    Creates a new workspace based on the method arguments

    Parameters
    ----------
    subscription_id : string
      Account subscription id (guid format).
    resource_group_name : TYPE
      Name of the resource group that will be created.
    workspace_name : TYPE
      Name of the workspace.
    auth : TYPE
      Authentification object (ex servcice principal from the AuthenticationService).

    Returns
    -------
    ws : azure.core.Workspace
      Newly created workspace or throws exception

    """
    self.log.p('Creating new workspace with resource group name `{}` and workspace name `{}`'.format(resource_group_name, workspace_name))
    ws = Workspace.create(
      auth=auth,
      subscription_id=subscription_id,
      resource_group=resource_group_name,
      name=workspace_name,
      create_resource_group=True,
      location=location
      )
    self.log.p('Done creating workspace', show_time=True)
    return ws
  
  def create_from_config(self, path_config, auth):
    """
    Creates a workspace from a config file

    Parameters
    ----------
    path_config : str
      Path to config file.
    auth : ServicePrincipalAuthentication
      Authentication proviced.

    Returns
    -------
    ws : Workspace
      Newly created workspace or throws exception.

    """
    self.log.p('Creating new workspace from config')
    ws = Workspace.from_config(path=path_config, auth=auth)
    self.log.p('Done creating new workspace from config')
    return ws
  
  def delete(self, ws, delete_dependent_resources=True, no_wait=False):
    """
    Deletes specified workspace instance

    Parameters
    ----------
    ws : Workspace
      Workspace instance.
    delete_dependent_resources : boolean, optional
      Whether to delete
            resources associated with the workspace, i.e., container registry, storage account, key vault, and
            application insights. The default is False. Set to True to delete these resources.. The default is True.
    no_wait : TYPE, optional
      Wheather to wait or not for deletion to complete. The default is False.

    Returns
    -------
    None if successfull or throws exception

    """
    ws.delete(
      delete_dependent_resources=delete_dependent_resources, 
      no_wait=no_wait
      )
    return
  
  def get(self, auth, subscription_id, resource_group_name, workspace_name):
    ws = Workspace.get(
      auth=auth, 
      subscription_id=subscription_id,
      resource_group=resource_group_name,
      name=workspace_name
      )
    return ws
  
  def list_workspaces(self, auth, subscription_id, resource_group_name=None):
    """
    This method queries subscription_id and returns a dict of active workspaces

    Parameters
    ----------
    auth : Service Principal
      Authentication object.
    subscription_id : string
      AzureML subscription id.

    Returns
    -------
    dct : dictionary
      Dictinary containing the list of active workspaces in the provided subscription.

    """
    return Workspace.list(
      auth=auth, 
      subscription_id=subscription_id,
      resource_group=resource_group_name
      )
  

class ModelService(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def create(self, workspace, name, version=1, tags=None, properties=None, model_framework=None, **kwargs):
    """
    Creates a new model based on the provided arguments

    Parameters
    ----------
    workspace : azure.core.Workspace
      Workspace used to register the model container.
    name : string, optional
      Model name.
    tags : list, optional
      An optional list of tags used to filter returned results. Results are filtered based on the provided list, searching by either 'key' or '[key, value]'. Ex. ['key', ['key2', 'key2 value']]. The default is None.
    properties : list, optional
      An optional list of properties used to filter returned results. Results are filtered based on the provided list, searching by either 'key' or '[key, value]'. Ex. ['key', ['key2', 'key2 value']]. The default is None.
    version : integer
      Model version.
    model_framework : string, optional
      Framework used to train the model. The default is None.
    **kwargs : TYPE
      DESCRIPTION.

    Returns
    -------
    azureml.core.model
      Instance of azureml model.

    """
    return Model(
      workspace=workspace, 
      name=name, 
      tags=tags, 
      properties=properties,
      version=version, 
      model_framework=model_framework,
      kwargs=kwargs
      )
  
  def register(self, workspace, model_path, model_name, description, tags=None, properties=None, model_framework=None):
    """
    This method handles the model container registration. You should perceive this "model" as a container/object that will be deployed on the workspace togheter with one or more assets

    Parameters
    ----------
    workspace : azureml.core workspace
      Workspace used to deploy the current model.
    model_path : string
      Path to assets: either a specific file (saved model pkl/h5/etc) or path to a folder that contain assets (model1.h5 + model2.pkl + model3.pb + data1.pkl + data2.csv + etc). 
    model_name : string
      Model name.
    description : string
      Description of the model.
    tags : list, optional
      An optional list of tags used to filter returned results. Results are filtered based on the provided list, searching by either 'key' or '[key, value]'. Ex. ['key', ['key2', 'key2 value']]. The default is None.
    properties : list, optional
      An optional list of properties used to filter returned results. Results are filtered based on the provided list, searching by either 'key' or '[key, value]'. Ex. ['key', ['key2', 'key2 value']]. The default is None.
    model_framework : TYPE, optional
      Framework used to develop/train the model. The default is None.

    Returns
    -------
    model : azureml.core Model
      Model container registered on the specified workspace.

    """
    assert os.path.exists(model_path)
    if os.path.isdir(model_path):
      assert len(os.listdir(model_path)) > 0
      
    model = Model.register(
      workspace=workspace, 
      model_path=model_path, 
      model_name=model_name, 
      tags=tags, 
      properties=properties, 
      description=description,
      model_framework=model_framework)
    return model
  
  def create_inference_config(self, entry_script_path, environment, source_directory=None):
    """
    This method creates inference config object based on the entry script and environment

    Parameters
    ----------
    entry_script_path : str
      Path to the script that will be used to make inference.
    environment : azureml environment
      Environment that will be used to make inference.

    Returns
    -------
    InferenceConfig
      Inference config object that can be later used to deploy/update a specific ACI container.

    """
    if source_directory:
      assert os.path.exists(os.path.join(source_directory, entry_script_path))
    else:
      assert os.path.exists(entry_script_path)
    return InferenceConfig(
      entry_script=entry_script_path, 
      environment=environment,
      source_directory=source_directory
      )

  def deploy(self, workspace, name, models, inference_config, deployment_config):
    """
    This method deploys a webservice as a real-time endpoint that can be used to make inference request for the models provided

    Parameters
    ----------
    workspace : azureml workspace
      A Workspace object to associate the Webservice with.
    name : str
      The name to give the deployed service. Must be unique to the workspace, only consist of lowercase letters, numbers, or dashes, start with a letter, and be between 3 and 32 characters long..
    models : list
      A list of model objects. Can be an empty list.
    inference_config : azureml inferenceconfig
      An InferenceConfig object used to determine required model properties.
    deployment_config : azureml WebserviceDeploymentConfiguration
      A WebserviceDeploymentConfiguration used to configure the webservice

    Returns
    -------
    None.

    """
    service = Model.deploy(
      workspace=workspace,
      name=name,
      models=models,
      inference_config=inference_config,
      deployment_config=deployment_config
      )
    service.wait_for_deployment(show_output=True)
    return service


class EnvironmentService(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def create_from_conda_specification(self, workspace, env_name, file_path):
    """
    Create environment object from an environment specification YAML file.
    Ex:
    name: inference_environment
    dependencies:
      - python=3.6.2
      - tensorflow>=2.0.0
      - pip:
        - azureml-defaults>=1.0.45
        
    Parameters
    ----------
    workspace : azureml.core workspace
      Current workspace used to deploy the model.
    env_name : string
      Conda environment name.
    file_path : TYPE
      Path to YAML config file.
      
    Returns
    -------
    conda_env : azureml.core Environemnt
      Registered environment container.

    """
    assert os.path.exists(file_path)
    assert file_path.endswith('.yaml'), 'Provided config should be yaml format'
    
    conda_env = Environment.from_conda_specification(
      name=env_name, 
      file_path=file_path
      )
    env = conda_env.register(workspace=workspace)
    return env
  
  def get(self, workspace, name, version=None):
    """
    Gets the environment specified by workspace and name

    Parameters
    ----------
    workspace : azureml workspace
      Workspace container.
    name : string
      Name of environment.
    version: string
      Environment version
      
    Returns
    -------
    Environment
      Returns the specified environment.

    """
    return Environment.get(workspace=workspace, name=name, version=version)
  
  def list_environments(self, workspace):
    """
    This method returns all the environments associated with a specific workspace

    Parameters
    ----------
    workspace : azureml workspace

    Returns
    -------
    Dict
      Dictinoary containing environments created in the specified workspace.

    """
    return Environment.list(workspace=workspace)


class AciService(BaseDecentrAIObject):
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    return
  
  def deploy_configuration(self, cpu_cores, memory_gb):
    """
    This method creates a configuration object with hardware specifications

    Parameters
    ----------
    cpu_cores : int
      Number of cpu cores.
    memory_gb : int
      Memory size.

    Returns
    -------
    service : AciServiceDeploymentConfiguration
      Config object with hardware resources.

    """
    service = AciWebservice.deploy_configuration(
      cpu_cores=cpu_cores, 
      memory_gb=memory_gb
      )
    return service
  
  def list_webservices(self, workspace):
    """
    List all the webservices associated with the provided workspace

    Parameters
    ----------
    workspace : azureml workspace
    Returns
    -------
    List of 
      List with all webservices associated with the provided workspace.

    """
    return AciWebservice.list(workspace=workspace)
  
  def delete(self, workspace, service_name):
    """
    This method deletes specified aci_service from the specified workspace) 

    Parameters
    ----------
    workspace : azureml workspace
      Workspace used to delete the aci webservice.
    service_name : str
      Service to be deleted

    Returns
    -------
    None.

    """
    try:
      service = self.get(workspace=workspace, service_name=service_name)
      service.delete()
    except Exception as e:
      self.log.p('There was an exception while deleting aci webservice: {}'.format(str(e)))
    return
  
  def get(self, workspace, service_name):
    """
    This method retrieves aci_service from the specified workspace

    Parameters
    ----------
    workspace : azureml workspace
      Workspace used to delete the aci webservice.
    service_name : str
      Service to be identified

    Returns
    -------
    AciWebservice.

    """
    return Webservice(workspace=workspace, name=service_name)


class AzureManager(BaseDecentrAIObject):
  def __init__(self, subscription_id, tenant_id, client_id, client_secret, **kwargs):
    """
    

    Parameters
    ----------
    subscription_id : String
      String containing Azure Subscription ID.
    tenant_id : String 
      String containing Azure Tenant ID.
    client_id : String
      String containing Azure Client ID.
    client_secret : String
      String containing Azure Client Secret.
    **kwargs : Dict
      BaseDecentrAIObject kwargs.

    Returns
    -------
    None.

    """
    self.__version__ = __version__
    self.version = __version__    
    self.__name__ = "AzureMLManager"
    
    super().__init__(**kwargs)
    self.subscription_id = subscription_id
    self.tenant_id = tenant_id
    self.client_id = client_id
    self.client_secret = client_secret
    
    self._auth()
    self.wss = WorkspaceService(log=self.log)
    self.envs = EnvironmentService(log=self.log)
    self.acis = AciService(log=self.log)
    self.ms = ModelService(log=self.log)
    return

  def _auth(self):
    auth = AuthenticationService(log=self.log)
    self.auth_sp = auth.service_principal_authentication(
      tenant_id=self.tenant_id,
      service_principal_id=self.client_id,
      service_principal_password=self.client_secret
      )
    return
  
  def _get_workspace(self, resource_group_name, workspace_name):
    ws = None
    try:
      dct_ws = self.wss.list_workspaces(
        auth=self.auth_sp, 
        subscription_id=self.subscription_id
        )
      if workspace_name in dct_ws:
        ws = self.wss.get(
          auth=self.auth_sp,
          subscription_id=self.subscription_id, 
          resource_group_name=resource_group_name, 
          workspace_name=workspace_name
          )
    except:
      self.log.p('Resource group does not exists')
    return ws
  
  def _create_workspace(self, resource_group_name, workspace_name):
    dct_ws = {}
    try:
      dct_ws = self.wss.list_workspaces(
        auth=self.auth_sp, 
        subscription_id=self.subscription_id,
        resource_group=resource_group_name
        )
    except:
      pass
    assert workspace_name not in dct_ws, 'Workspace name you provided {} already exist in the resource group {}'.format(workspace_name, resource_group_name)
   
    ws = self.wss.create(
      subscription_id=self.subscription_id,
      resource_group_name=resource_group_name,
      workspace_name=workspace_name,
      auth=self.auth_sp
      )
    return ws
  
  def _get_conda_env(self, workspace, inference_env_name):
    dct_envs = self.envs.list_environments(workspace=workspace)
    env_conda = None
    if inference_env_name in dct_envs:
      env_conda = self.envs.get(workspace=workspace, name=inference_env_name)
    return env_conda
  
  def _create_conda_env(self, workspace, inference_env_name, conda_config_path):
    dct_envs = self.envs.list_environments(workspace=workspace)
    assert inference_env_name not in dct_envs, 'Inference env name provided already exists in the workspace: {}'.format(inference_env_name)
    
    env_conda = self.envs.create_from_conda_specification(
      workspace=workspace, 
      env_name=inference_env_name, 
      file_path=conda_config_path
      )
    return env_conda
  
  def new_webservice(self, resource_group_name, workspace_name, 
                     inference_env_name, conda_config_path, 
                     scoring_path, webservice_name, cpu_cores, memory_gb, 
                     model_path, model_name, model_tags=None, model_description=None,
                     scoring_source_directory=None):
    """
    This method creates a new webservice in Azure, by automatically allocating hardware and creating the inference environment.

    Parameters
    ----------
    resource_group_name : String
      The name of the existing or future Azure resource group.
    workspace_name : String
      The name of the existing or future Azure workspace.
    cpu_cores : Int
      The number of CPU cores to allocate to the future Azure container.
    memory_gb : Int
      The memory to allocate to the future Azure Container.
    inference_env_name : String
      The name of the (inference) environment.
    conda_config_path : String
      Path to the .yaml conda config file.
    scoring_path : String
      The name of the scoring script (.py script)
    model_path : String
      Path to the model.
    model_name : String
      Model name.
    model_tags : Dict
      Dictionary containing tags describing the model.
    model_description : String
      Description of the model.
    webservice_name : String
      The name for the final created webservice.
    scoring_source_directory: String
      The path to entire data structure (folder) to be uploaded into Azure
    
    Returns
    -------
    service : Azure Webservice
      Service that can be accessed and used to make predictions.

    """
    t_start = time()
    self.log.p('Deploying new webservice')
    self.log.p(' * Resource Group Name: ' + resource_group_name)
    self.log.p(' * Workspace Name: ' + workspace_name)
    self.log.p(' * Inference Env Name: ' + inference_env_name)
    self.log.p(' * Webservice Name: ' + webservice_name)
    #1. get or create workspace & resource group
    ws = self._get_workspace(resource_group_name, workspace_name)
    if not ws:
      ws = self._create_workspace(resource_group_name, workspace_name)
    self.log.p('Workspace loaded', show_time=True)
    
    #2. get or create the required conda environmnet for model
    env_conda = self._get_conda_env(ws, inference_env_name)
    if not env_conda:
      env_conda = self._create_conda_env(ws, inference_env_name, conda_config_path)
    self.log.p('Conda environment loaded', show_time=True)
    #3. deploy resources
    aci_config = self.acis.deploy_configuration(cpu_cores, memory_gb)
    self.log.p('Deploy configuration loaded', show_time=True)
    
    #4. create inference config based on scoring script and environment
    inf_conf = self.ms.create_inference_config(
      scoring_path, 
      env_conda, 
      scoring_source_directory
      )
    self.log.p('Inference configuration loaded', show_time=True)
    
    #5. register the assets (models, etc) needed to make inference
    model = self.ms.register(
      workspace=ws, 
      model_path=model_path,
      model_name=model_name,
      tags=model_tags,
      description=model_description
      )
    self.log.p('Model registered', show_time=True)
    
    #6. deploy
    service = self.ms.deploy(
      workspace=ws,
      name=webservice_name,
      models=[model],
      inference_config=inf_conf,
      deployment_config=aci_config
      )
    self.log.p('Model deployed', show_time=True)
    t_stop = time()
    self.log.p('Total time: {}'.format(t_stop - t_start))
    return service
  
  def update_webservice(self, resource_group_name, workspace_name, inference_env_name, 
                        webservice_name, scoring_path, 
                        model_path, model_name, model_tags=None, model_description=None,
                        scoring_source_directory=None):
    """
    This method updates and existing webservice by uploading a new model and a new scoring script.

    Parameters
    ----------
    resource_group_name : String
      The name of the existing or future Azure resource group.
    workspace_name : String
      The name of the existing or future Azure workspace.
    inference_env_name : String
      The name of the (inference) environment.
    scoring_path : String
      The name of the scoring script (.py script)
    model_path : String
      Path to the model.
    model_name : String
      Model name.
    model_tags : Dict
      Dictionary containing tags describing the model.
    model_description : String
      Description of the model.
    webservice_name : String
      The name for the final created webservice.
    scoring_source_directory: String
      The path to entire data structure (folder) to be uploaded into Azure

    Returns
    -------
    webservice : Azure Webservice
      New service updated and ready to be used for inference.

    """
    t_start = time()
    self.log.p('Updating webservice:')
    self.log.p(' * Resource Group Name: ' + resource_group_name)
    self.log.p(' * Workspace Name: ' + workspace_name)
    self.log.p(' * Inference Env Name: ' + inference_env_name)
    self.log.p(' * Webservice Name: ' + webservice_name)
    #1. check and get the workspace
    dct_ws = self.wss.list_workspaces(auth=self.auth_sp, subscription_id=self.subscription_id)
    assert workspace_name in dct_ws, 'The workspace you provided does not exist on this subscription!'
    workspace = self._get_workspace(resource_group_name, workspace_name)
    self.log.p('Workspace loaded', show_time=True)
    
    #2. check and get the (inference) environment
    dct_envs = self.envs.list_environments(workspace=workspace)
    assert inference_env_name in dct_envs, 'The inference environment you provided does not exist on this workspace'
    env_conda = self._get_conda_env(workspace, inference_env_name)
    self.log.p('Environment loaded', show_time=True)
    
    #3. check and get the existing webservice
    webservice = self.acis.get(workspace, webservice_name)
    assert webservice is not None, 'Service {} does not exist in the current workspace {}'.format(webservice_name, workspace_name)
    self.log.p('Webservice loaded', show_time=True)
    
    #4. creating new inference config with the new score.py
    new_inference_config = self.ms.create_inference_config(
      scoring_path, 
      env_conda, 
      scoring_source_directory)
    self.log.p('New inference config created', show_time=True)
    
    #5. creating new model
    new_model = self.ms.register(
      workspace=workspace,
      model_path=model_path,
      model_name=model_name,
      tags=model_tags,
      description=model_description
      )
    self.log.p('New model registered', show_time=True)
    
    webservice.update(models=[new_model], inference_config=new_inference_config)
    webservice.wait_for_deployment(show_output=True)
    t_stop = time()
    self.log.p('Total time: {}'.format(t_stop - t_start))
    return webservice
    


if __name__ == '__main__':

  def test():
    headers = {'Content-Type': 'application/json'}
    payload = json.dumps({"text": "am fost la un restaurant cu muzica placuta si personal dragut. ciorba lor de legume era delicioasa. fructele lor de mare erau din alta lume. caviarul a fost delicios. am uitat sa mentionez ca restaurantul era pe Calea Dorobanti"})
    response = requests.post(service.scoring_uri, data=payload, headers=headers)
    log.p('The service can be called at: {}'.format(service.scoring_uri))
    log.p('Service responded with code: {}'.format(response.status_code))
    log.p('Service responded in: {}'.format(response.elapsed))
    log.p('Service response: {}'.format(response.json()))
    return

  import json
  import requests
  from logger import Logger

  log = Logger(lib_name='AZRML', config_file='config.txt')
  path_data = os.path.join(log.get_data_folder(), 'azureml')
  
  AZURE_SUBSCRIPTION_ID = '21862221-de06-45b9-b99b-94e0a3c69193'
  AZURE_TENANT_ID       = '6f2b3c9b-f251-47cd-add5-3c3bca51b59b'
  AZURE_CLIENT_ID       = '5e548f9b-4884-40cd-98b0-c18676a29f62'
  AZURE_CLIENT_SECRET   = 'FR1HX35VS_ssGmiYsBO5..ogl1a.83g0cx'

  resource_group_name       = 'HODS3'
  workspace_name            = 'Lecture8'
  inference_env_name        = 'hods_env'
  cpu_cores                 = 1
  memory_gb                 = 1
  conda_config_path         = os.path.join(path_data, 'assets', 'config.yaml')
  model_path                = os.path.join(path_data, 'models')
  model_name                = 'restocracy'
  model_tags                = {'restocracy': 'demo'}
  model_description         = 'Model trained on restocracy reviews'
  webservice_name           = 'restocracyapi'
  scoring_source_directory  = None
  scoring_path              = os.path.join(path_data, 'test1', 'score.py')

  azure_mng = AzureManager(
    log=log, 
    subscription_id=AZURE_SUBSCRIPTION_ID,
    tenant_id=AZURE_TENANT_ID,
    client_id=AZURE_CLIENT_ID,
    client_secret=AZURE_CLIENT_SECRET
    )
  
  #Test case 1:
  #We have a new model and we need to fastly serve it in Azure
  # For that we are going to create a new webservice
  # Create webservice
  service = azure_mng.new_webservice(
    resource_group_name=resource_group_name, 
    workspace_name=workspace_name, 
    inference_env_name=inference_env_name, 
    conda_config_path=conda_config_path, 
    webservice_name=webservice_name,
    cpu_cores=cpu_cores, 
    memory_gb=memory_gb, 
    scoring_path=scoring_path,
    scoring_source_directory=scoring_source_directory,
    model_path=model_path, 
    model_name=model_name, 
    model_tags=model_tags, 
    model_description=model_description
    )
  test()
  
  #Test case 2:
  #After some days we decide to update the scoring script in order to respond with a json containing the model version.
  #So we do some updates on our scoring script and after that we are updating our webservice
  # Update webservice
  scoring_path = os.path.join(path_data, 'test2', 'score.py')
  service = azure_mng.update_webservice(
    resource_group_name=resource_group_name, 
    workspace_name=workspace_name, 
    inference_env_name=inference_env_name, 
    webservice_name=webservice_name,
    scoring_path=scoring_path,
    scoring_source_directory=scoring_source_directory,
    model_path=model_path,
    model_name=model_name,
    model_tags=model_tags,
    model_description=model_description
    )
  test()
  
  #Test case 3:
  #After a week of using the version 2 of our scoring/model, we decide to include logger in our deployment.
  #The key aspect here is to specify that you have an entire directory that you want to upload
  #The scoring script is still inside that directory
  
  scoring_source_directory = os.path.join(path_data, 'test3')
  scoring_path = 'score.py'
  
  # Update webservice
  service = azure_mng.update_webservice(
    resource_group_name=resource_group_name, 
    workspace_name=workspace_name, 
    inference_env_name=inference_env_name, 
    webservice_name=webservice_name,
    scoring_path=scoring_path,
    scoring_source_directory=scoring_source_directory,
    model_path=model_path,
    model_name=model_name,
    model_tags=model_tags,
    model_description=model_description
    )
  test()

  # service.get_logs()




#global dependencies
import json
import re
import abc
import traceback
from collections import OrderedDict
from multiprocessing import Process, parent_process
from time import time, sleep

from ratio1.ipfs import R1FSEngine

#local dependencies
from naeural_core import constants as ct
from naeural_core import Logger
from naeural_core.local_libraries import _ConfigHandlerMixin
from naeural_core.serving.mixins_base import (
  _InferenceUtilsMixin,
)
from naeural_core.utils.system_shared_memory import NumpySharedMemory, replace_shm_with_ndarray, replace_ndarray_with_shm

# from naeural_core.serving.base.serving_utils import CommEngine


_CONFIG = {

  "SERVING_TIMERS_IDLE_DUMP"              : 3601,
  "SERVING_TIMERS_IDLE_DUMP_DEFAULT"      : 1801,
  "SERVING_TIMERS_PREDICT_DUMP"           : 901,
  "SERVING_TIMERS_PREDICT_DUMP_DEFAULT"   : 601,
  
  "R1FS_ENABLED"                          : False,

  "CLOSE_IF_UNUSED"                       : False,
  
  "MAX_WAIT_TIME"                         : 5,
  
  "MODEL_ZOO_CONFIG"                      : {},
  "CUSTOM_DOWNLOADABLE_MODEL"             : False,
  "CUSTOM_DOWNLOADABLE_MODEL_URL"         : None,
  "PICKED_INPUT"                          : "IMG",
  "RUNS_ON_EMPTY_INPUT"                   : False,
  "MODEL_INSTANCE_ID"                     : None,
  

  ## SERIALIZATION & PERSISTENCE
  'LOAD_PREVIOUS_SERIALIZATION'   : False,
  'SERIALIZATION_SIGNATURE'       : None,
  ## END SERIALIZATION & PERSISTENCE
  
  
  
  'VALIDATION_RULES' : {

  },
}

class ModelServingProcess(
  _ConfigHandlerMixin,
  _InferenceUtilsMixin,
  Process,
):
  """
  This is the basic model serving process class. 
  
  The following rules must be complied for subclassing:
    
    `startup`         : define a one time model setup method
    
    `on_init`         : define a one time model setup method executed after `startup`
    
    `pre_process`     : define a function that prepares (if needed) input for `_predict`. Just
                        return the inputs if not required.
    
    `predict`         : define a predict function that receives `_pre_process_inputs` result
    
    `post_process`    : define a function that post-processes (if needed) the `_predict`
                        output. Just return the function input if no post-proc required
  """

  CONFIG = _CONFIG
  __metaclass__ = abc.ABCMeta

  def __init__(self, 
               server_name : str, 
               comm_eng, 
               inprocess: bool,
               default_config,
               version : str = "0.0.0",
               upstream_config=None,
               full_debug=False, 
               log=None,
               try_predict=True,
               environment_variables=None,
               npy_shm_kwargs=None,
               comm_method=None,
               **kwargs
               ):
    self._done = False
    self.log = log
    self.inprocess = inprocess
    self._full_debug = full_debug
    self._environment_variables = environment_variables
    self.server_name = server_name
    
    # `predict_server_name` initially defaults to actual server and can be overwritten
    # in `predict` with a given covered server    
    self.predict_server_name = server_name 
    # this must be correctly configured by each individual serving process
    # and is redundant for a lot of models
    self._has_second_stage_classifier = False    
    # end `predict_server_name` config
    
    _name = self.server_name if len(self.server_name) < 10 else Logger.name_abbreviation(self.server_name)
    self.__prefix_name = _name
    self.prefix = '[' + _name + ']'
    self.comm_eng = comm_eng
    self.__version__ = version
    self.version = self.__version__
    self._default_config = default_config
    self._upstream_config = upstream_config or {}
    self.config_model = None
    self.r1fs : R1FSEngine = None
    self._startup_failure = False
    self._inprocess_startup_done = False
    self._displayed_first_predict_coverage = False
    self._parents = self.get_server_parents()
    self.TRY_PREDICT = try_predict
    self._stream_index_mapping = None

    self.inputs = None
    self.prep_inputs = None
    self.preds = None
    self.npy_shm_kwargs = npy_shm_kwargs
    self.npy_shm = None
    self.comm_method = comm_method

    super(ModelServingProcess, self).__init__()
    if self.inprocess:
      self.inprocess_startup()
    return
  
  @property
  def _cache_folder(self):
    return ct.CACHE_SERVING
    
  @property
  def plugin_id(self):
    return self.sanitize_name(self.server_name)
  
  @property
  def json(self):
    return json

  @property
  def has_second_stage_classifier(self):
    return self._has_second_stage_classifier and self.server_name == self.predict_server_name

  @property
  def serving_timers_idle_dump(self):
    return max(self.cfg_serving_timers_idle_dump, self.cfg_serving_timers_idle_dump_default)

  @property
  def serving_timers_predict_dump(self):
    return max(self.cfg_serving_timers_predict_dump, self.cfg_serving_timers_predict_dump_default)

  @property
  def close_if_unused(self):
    return self.cfg_close_if_unused if self.ready_cfg_handlers else False

  def sleep(self, sec):
    sleep(sec)

  def get_serving_param(self, key, inference_upstream_config):
    val = inference_upstream_config.get(key, None)
    if val is None:
      val = self.config_model[key]
    return val
  
  def download(self, url, fn):
    kwargs = self.cfg_model_zoo_config 
    return self.log.maybe_download_model(
      url=url,
      model_file=fn,
      **kwargs,
    )

  def get_server_parents(self):
    parents = self.__class__.mro()
    parent_servers = []
    for parent in parents:
      parent_name = parent.__name__
      subnames = re.findall('[A-Z][^A-Z]*', parent_name)
      parent_server = "_".join(subnames).upper()
      if parent_server == 'MODEL_SERVING_PROCESS':
        break
      parent_servers.append(parent_server)
    return parent_servers[1:]
  
  def _start_timer(self, tmr):
    self.log.start_timer(self.server_name + '_' + tmr)
    return
  
  def _stop_timer(self, tmr, periodic=False):
    self.log.stop_timer(self.server_name + '_' + tmr, periodic=periodic)
    return

  def start_timer(self, tmr):
    self.log.start_timer(self.server_name + '_' + tmr)
    return
  
  def stop_timer(self, tmr, periodic=False):
    self.log.stop_timer(self.server_name + '_' + tmr, periodic=periodic)
    return

  def end_timer(self, tmr, periodic=False):
    self.log.stop_timer(self.server_name + '_' + tmr, periodic=periodic)
    return


  def get_config(self):
    return self.config_model
  
  def _get_message(self, wait=False):
    payload, msg_type, msg_data, msg_server_name, msg_id, msg_ts = None, None, None, None, None, None
    if not wait:
      sleep(0.0001)
    if self._full_debug:
      # sleep(1)
      self.P("Running poll...")
    res = self.comm_eng.poll(0.5)
    if wait or res:
      if self._full_debug and res:
        self.P("{} has data and will `recv`".format(self.__class__.__name__), color='y')
      payload = self.comm_eng.recv()
        
    if payload is not None:
      if self._full_debug:
        self.P(f'Received payload with keys: {payload.keys()}')
        self.P(f'payload[`SERVER_NAME`]={payload.get("SERVER_NAME")}')
      msg_type = payload['TYPE']
      msg_data = replace_shm_with_ndarray(payload['DATA'], self.npy_shm, debug=self._full_debug)
      msg_server_name = payload['SERVER_NAME']
      msg_id = payload['CMD_ID']
      msg_ts = payload['TIMESTAMP']
      elapsed = self.time() - msg_ts
      if self._full_debug:
        self.P("  {}:{} received comand {}:{} and data type '{}'[Elapsed: {}s]".format(
          self.__class__.__name__, msg_server_name, msg_type, msg_id,
          msg_data.__class__.__name__, round(elapsed, 3)), color='y'
        )
    return msg_type, msg_data, msg_server_name, msg_id, msg_ts
  
  
  def get_name(self):
    return self.__class__.__name__, self.server_name, self.name
  
  
  def _send_message(self, command, inputs, send_timers=False, cmd_id=None):
    timers_val = self.log.export_timers_section() if send_timers else None
    if timers_val is not None:
      self.P("Sending timers (len {})".format(len(timers_val[0])))
    dct_msg = {
      'TYPE'    : command,
      'DATA'    : inputs,
      'TIMERS'  : timers_val,
      'CMD_ID'  : cmd_id,
      'TIMESTAMP': self.time(),
    }
    if self._full_debug:
      self.P("Sending '{}' for cmd_id:{} to manager".format(command, cmd_id))
    self.comm_eng.send(dct_msg)
    if self._full_debug:
      self.P("  '{}' for cmd_id:{} sent to manager".format(command, cmd_id))
    return
  
  def _dump_error(self, data):
    str_err = traceback.format_exc()
    fn_input = self.log.save_pickle(
      data=data, 
      fn='{}_predict_error.pkl'.format(self.__class__.__name__),
      folder='output',
      use_prefix=True,
      )
    fn_text = self.log.get_output_folder() + "/{}_{}_predict_error.txt".format(
      self.log.file_prefix,
      self.__class__.__name__,
      )
    with open(fn_text, "wt") as fp:
      fp.write(str_err)
    return fn_input, fn_text
  
  def _create_logger(self):
    self.log = Logger(
      lib_name=self.__prefix_name,
      base_folder='.',
      app_folder=ct.LOCAL_CACHE,
      TF_KERAS=False,
      default_color=ct.COLORS.SERVING,
    )      
    return

  def create_numpy_shared_memory(self):
    """
    Create a numpy shared memory object for accessing data loaded by the serving manager.
    Returns
    -------

    """
    if self.npy_shm_kwargs is not None and self.comm_method != 'pipe':
      self.npy_shm = NumpySharedMemory(
        **self.npy_shm_kwargs,
        create=False,
        log=self.log
      )
    return

  def inprocess_startup(self):
    if self._inprocess_startup_done:
      self.P("inprocess_startup` already called. Bypassing.", color='y')
      return
    if self.inprocess:
      # now we prepare config
      self._setup_config()
      _ = self._startup() 
      self._inprocess_startup_done = True
    else:
      raise ValueError('`inprocess_startup` called without actual inprocess serving process!')
    return

  def _prepack_results(self, results):
    results_dict = OrderedDict()
    # TODO: should the additional be added after the rest?
    meta = OrderedDict()
    additional_meta = self.get_additional_metadata()
    if isinstance(additional_meta, dict):
      meta.update(additional_meta)
    else:
      meta['ADDITIONAL_META'] = additional_meta
    # endif additional meta

    if results is None or len(results) == 0:
      meta[ct.ERROR] = 'No inputs received for inference'
      results = [[]]

    meta[ct.SYSTEM_TIME] = self.log.now_str(nice_print=True, short=False)
    meta[ct.VER] = self.__version__
    meta['PICKED_INPUT'] = self.cfg_picked_input
    results_dict['INFERENCES_META'] = meta
    return results_dict, results

  def _pack_results(self, results):
    results_dict, results = self._prepack_results(results)
    if len(self._stream_index_mapping) == 0:
      results_dict[ct.INFERENCES] = results
    else:
      results_dict[ct.INFERENCES] = []
      for stream_name, indexes in self._stream_index_mapping.items():
        crt_stream_results = []
        for i in indexes:
          crt_stream_results.append(results[i])
        results_dict[ct.INFERENCES].append(crt_stream_results)
      #endfor
    #endif
    
    return results_dict

  def _input_validator(self):
    pass ###TODO

  def _pick_inference_inputs(self, inputs, data_type):
    """
    Inputs standard:
    [
      {
        "STREAM_NAME" : "xxxxx",
        "STREAM_METADATA" : {"k1" : v1, "k2" : v2, ...},
        "INPUTS" : [
          {
            "IMG" : np.ndarray,
            "STRUCT_DATA" : None,
            "INIT_DATA" : None,
            "METADATA" : {"k1" : v1, "k2" : v2, ...},
            "TYPE" : "IMG"
          }
        ],
        "SERVING_PARAMS" : {...},
      }
    ]

    - `inputs` is a batch (list) of stream collected data (because the inference is done batchified on all streams that use the same model)
    - each element of `inputs` is a dictionary having the following keys:
      * STREAM_NAME
      * STREAM_METADATA
      * INPUTS
      * SERVING_PARAMS

    - `inputs[i]['SERVING_PARAMS']` defines the upstream configuration for the current serving plugin 

    - `inputs[i]['INPUTS']` is also a list (a stream can collect data from multiple sources - multi-modal or meta-stream).
      Each element in this list is basically a sub-stream.
      When we have to deal with simple cases (a stream collects a video frame and that'a all), `inputs[i]['INPUTS']` is a list of 1 element.

    - each element of `inputs[i]['INPUTS']` is a dictionary having the following keys:
      * IMG - if the sub-stream returned an image
      * STRUCT_DATA - if the sub-stream returned a chuck of structured data
      * INIT_DATA - if the sub-stream returned some initial data
      * METADATA - sub-stream metadata
      * TYPE: "IMG" or "STRUCT_DATA" (they are mutual exclusive and cannot co-exist, i.e. a sub-stream returns either IMG, or STRUCT_DATA)

    """

    dct_picked = {
      'DATA' : [],
      'INIT_DATA' : [],
      'SERVING_PARAMS' : [],
      'STREAM_NAME' : [],
      'EMPTY' : False
    }

    for stream_input in inputs:
      stream_name = stream_input.get('STREAM_NAME')
      if stream_name is None:
        continue
      self._stream_index_mapping[stream_name] = []
      lst_sub_stream_inputs = stream_input.get('INPUTS') or []
      _serving_params = stream_input.get('SERVING_PARAMS', {})

      for _input in lst_sub_stream_inputs:
        input_type = _input.get('TYPE')
        if input_type == data_type:
          picked_data = _input.get(data_type)
          if picked_data is None:
            continue
          dct_picked['DATA'].append(picked_data)
          dct_picked['SERVING_PARAMS'].append(_serving_params)
          dct_picked['INIT_DATA'].append(_input.get('INIT_DATA'))
          dct_picked['STREAM_NAME'].append(stream_name)
          self._stream_index_mapping[stream_name].append(len(dct_picked['DATA'])-1)
        #endif
      #endfor
    #endfor

    if len(dct_picked['STREAM_NAME']) == 0:
      dct_picked['EMPTY'] = True

    return dct_picked

  def maybe_log_phase(self, phase, start_time, done=True):
    if self._full_debug:
      prefix = 'Done' if done else 'Started'
      self.P(f'[DEBUG]{prefix} {phase} for {self.server_name}[Elapsed: {round(time() - start_time, 3)}]')
    return

  def __predict(self, inputs, **kwargs):
    """
    Parameters:
    -----------
    inputs: dict, mandatory
      pass
    """
    start_time = time()
    self.inputs = inputs
    if inputs.pop('EMPTY', False):
      outputs = []
    else:
      # PRE-PROCESS
      self.maybe_log_phase('_pre_process', start_time, done=False)
      self._start_timer(ct.TIMER_PRE_PROCESS)
      prep_inputs = self._pre_process(inputs)
      self.prep_inputs = prep_inputs
      self._stop_timer(ct.TIMER_PRE_PROCESS)
      self.maybe_log_phase('_pre_process', start_time)

      # INFERENCE/PREDICTION
      self.maybe_log_phase('_predict', start_time, done=False)
      self._start_timer(ct.TIMER_RUN_INFERENCE)
      preds = self._predict(prep_inputs)
      self.preds = preds  # TODO: review why this is here
      self._stop_timer(ct.TIMER_RUN_INFERENCE)
      self.maybe_log_phase('_predict', start_time)

      # POST-PROCESS
      self._start_timer(ct.TIMER_POST_PROCESS)
      outputs = self._post_process(preds)
      self._stop_timer(ct.TIMER_POST_PROCESS)
      self.maybe_log_phase('_post_process', start_time)
    #endif

    # TO DICT
    self._start_timer(ct.TIMER_PACK_RESULTS)
    results = self._pack_results(outputs)
    self._stop_timer(ct.TIMER_PACK_RESULTS)
    self.maybe_log_phase('_pack_results', start_time)

    return results
  
  
  def _in_process_predict(self, inputs, **kwargs):
    return self.__loop_predict(inputs, **kwargs)
  
  def __loop_predict(self, inputs, **kwargs):
    """
    Parameters:
    ----------
    inputs: list[object], mandatory
      List of all inference inputs.
      If the list contain dictionaries, it means that the inputs are collected and prepared in Orchestrator's main loop;
        thus they should be "picked" by `self._pick_inference_inputs`
      If the list contain other types, it means that the inputs are raw.
    """
    assert isinstance(inputs, list) and len(inputs) >= 1 or self.cfg_runs_on_empty_input

    self._start_timer(ct.TIMER_PREDICT)
    self._stream_index_mapping = {}
    self.kwargs = kwargs
    
    self.predict_server_name = kwargs.get('server_name', self.server_name)
    if not self._displayed_first_predict_coverage:
      self._displayed_first_predict_coverage = True
      if self.predict_server_name != self.server_name:
        self.P("Running '{}' for covered server '{}'".format(self.server_name, self.predict_server_name))

    dct_picked = {
      'DATA': inputs,
      'EMPTY': False
    }
    if isinstance(inputs[0], dict):
      dct_picked = self._pick_inference_inputs(inputs, self.cfg_picked_input)
    #endif
    if self.cfg_runs_on_empty_input:
      dct_picked['EMPTY'] = False
    # endif empty input allowed

    if self.TRY_PREDICT:
      try:
        results = self.__predict(dct_picked, **kwargs)
      except:
        res = self._dump_error(inputs)
        msg = "Error raised in {}.predict(). Dump prepared. Please see error detailed info here: {}".format(self.__class__.__name__, res)
        raise ValueError(msg)
    else:
      results = self.__predict(dct_picked, **kwargs)

    self._stop_timer(ct.TIMER_PREDICT)
    return results
  
  
  def P(self, msg, color=None, **kwargs):
    if self.inprocess:
      prefix = self.prefix + f'[{self.pid}] '
    else:
      prefix = f'[{self.pid}] '

    if color not in ['e', 'r']:
      color = ct.COLORS.SERVING
    self.log.P(prefix + msg, color=color, **kwargs)
    return
  
  def _predict_iter_debug_handle(self, curr_iter, n_thr=100):
    TEST = None # sleep, crash or None
    if curr_iter >= n_thr and isinstance(TEST, str):
      TEST = TEST.lower()      
      self.P("DEBUG predict iter {}: executing TEST='{}'".format(
        curr_iter, TEST), color='r'
      )
      if TEST == 'sleep':
        sleep_time = 120
        self.P("  Sleeping for {} sec to annoy the SM and kill the current process".format(
          sleep_time), color='r'
        )
        sleep(sleep_time)
      elif TEST == 'crash':
        raise ValueError("Exceptionen!")            
      else:
        self.P("  Unknown TEST='{}'".format(TEST), color='r')
      # end tests
    # end debug threshold
    return
  
  def run(self):
    if self.inprocess:
      raise ValueError("Something is wrong: `{}.run` called for a inprocess serving process".format(
        self.__class__.__name__))
    try:
      # all following code is "protected" in one big exception catching block, thus any
      # major exception will stop the process and announce the Serving Manager
      # so we do not necesarely need higher levels of exception handling unless 
      # those errors are minor and we can work around them. 
      
      self.__parent_process = parent_process()
      
      n_predicts = 0
      self._create_logger() # uses only server_name
      self.create_numpy_shared_memory()
      # prepare config
      self._setup_config() 
      
      # startup: finally this is where the model (should be) is created in theory
      self._startup_failure = True # we assume anything can go wrong
      self._start_timer('startup')
      _msg = self.__startup() 
      self.__on_init()
      self._stop_timer('startup')
      self._startup_failure = False # all passed
      
      
      tm_last_timers_dump = None
      if _msg is None:
        _msg = "{} ({}) `_startup` done.".format(self.server_name, self.name)
      self._send_message(command='READY', inputs=_msg)
      self.P("{} is waiting for commands...".format(self.server_name), color='g')
      must_send_timers = False
      while not self._done:
        cmd, data, server_name, cmd_id, ts = self._get_message(wait=False)
        # now we parse the commands
        if cmd is not None:
          if cmd.upper() == 'STOP':
            self.P("Received 'STOP' for {}.".format(self.server_name))
            self._done = True
          # end STOP
          elif cmd.upper() == 'PREDICT':
            if self._full_debug:
              self.P(f'[DEBUG]Received PREDICT for {self.server_name}[Elapsed: {round(self.time() - ts, 3)}]')
            res = self.__loop_predict(data, server_name=server_name)
            if self._full_debug:
              self.P(f'[DEBUG]Done PREDICT for {self.server_name}[Elapsed: {round(self.time() - ts, 3)}]')
            n_predicts += 1
            if tm_last_timers_dump is None or (time() - tm_last_timers_dump) > self.serving_timers_predict_dump:
              self.P("Predict timers dump time {}s reached. Dumping timers and sending to Serving Manager".format(self.serving_timers_predict_dump))
              tm_last_timers_dump = time()
              self.log.show_timers(color='d')
              must_send_timers = True
            # now we can write a run-each-n-predicts debug section
            self._predict_iter_debug_handle(curr_iter=n_predicts) # DEBUG method
            # send message with the result if managed so far
            self._send_message(
              command='RESULT', 
              inputs=res,
              send_timers=must_send_timers,
              cmd_id=cmd_id,
            )
            if self._full_debug:
              self.P(f'[DEBUG]Sent RESULT for {self.server_name}[Elapsed: {round(self.time() - ts, 3)}]')
            must_send_timers = False
          # end PREDICT
          elif cmd.upper() == 'GET_CONFIG':
            self.P("Sending config to the serving manager...")
            self._send_message(
              command='GET_CONFIG',
              inputs=self.get_config(),
              send_timers=False,
            )
          # end GET_CONFIG
          else:
            self.P("Received unknown command '{}'".format(cmd.upper()), color='r')
          # endif select cmd type
        else:
          # TODO: add on_idle method here
          # now we are on IDLE - no cmd received so we can do quick internal kitchen
          if tm_last_timers_dump is None or (time() - tm_last_timers_dump) > self.serving_timers_idle_dump:
            self.P("Idle timers dump {}s reached. Dumping timers".format(self.serving_timers_idle_dump))
            tm_last_timers_dump = time()
            self.log.show_timers(color='d')
            must_send_timers = True
        #endif command or none
        if self.__parent_process is not None and not self.__parent_process.is_alive():
          self.P("Parent process {} is dead. Stopping...".format(self.__parent_process.pid), color='r')
          self._done = True
        # end parent process check
      #endwhile main loop
      self.P("Gracefully closing Model Server '{}'".format(self.server_name), color='r')
    except:
      str_except = traceback.format_exc()
      if self._startup_failure:
        _error = 'STARTUP_FAILURE'
      else:
        _error = 'ABNORMAL_EXIT'
      self.P("Exception in {}:\n{}".format(
                self.__class__.__name__, 
                str_except
              ),
              color="r",
            )
      self._send_message(command=_error, inputs=str_except)
    #end try-except
    self.P("Closing Serving Process execution loop...")
    # TODO: cleanup stuff
    self.shutdown(from_parallel=True)
    self.log.show_timers()
    return
      
  def shutdown(self, from_parallel=False):
    if self.inprocess:
      self.P("Shutting down inprocess server...")
    elif not from_parallel:
      self.P("WARNING: called shutdown on a parallel process without explicit knowledge!")
    else:
      self.P("Shutting down parallel process!")
    if self.npy_shm is not None:
      self.npy_shm.shutdown()
    self._shutdown()
    return

  def __on_init(self):
    if self.cfg_r1fs_enabled:
      self.r1fs = R1FSEngine(
        logger=self.log,
      )
    self._on_init()
    return
  
  def __startup(self):
    return self._startup()    
  # START: SUCCLASS MUST IMPLEMENET
    
  
  def startup(self):
    """
    This method is called at startup. It can be used to prepare the model.
    """
    return
  
  def _startup(self):
    return self.startup()
    
  
  def on_init(self):
    """
    This method is called post startup. It can be used for additional prep if needed.
    """
    return
    
  def _on_init(self):
    self.on_init()
    return

  def get_additional_metadata(self):
    """
    This method is called to add metadata to what is sent to the serving manager.
    Returns
    -------

    dict
        The additional metadata to be sent to the serving manager
    """
    return {}

  @abc.abstractmethod
  def pre_process(self, inputs):
    """
    This method is called before the actual prediction. It can be used to prepare the inputs.
    The implementation is specific for each serving process.
    """
    raise NotImplementedError()
  
  def _pre_process(self, inputs):
    return self.pre_process(inputs)


  @abc.abstractmethod
  def predict(self, prep_inputs):
    """
    This method is called to perform the actual prediction. It can be used to perform the prediction.
    The implementation is specific for each serving process.
    """
    raise NotImplementedError()
      
  def _predict(self, prep_inputs):
    return self.predict(prep_inputs)


  @abc.abstractmethod
  def post_process(self, preds):
    """
    This method is called after the actual prediction. It can be used to post-process the predictions.
    The implementation is specific for each serving process.
    """
    raise NotImplementedError()
  
  def _post_process(self, preds):
    return self.post_process(preds)
  # END: SUCCLASS MUST IMPLEMENET


  # START: SUBCLASS CAN OVERWRITE

  def _shutdown(self):
    return

  def _setup_config(self):
    """
    if subclassed, mandatory call super()._setup_config() in first line
    """
    dct_cfg = None
    dct_env = self._environment_variables # here we can use self._environment_variables["SERVING_PROCESS_CONFIG"]
    if isinstance(dct_env, dict) and len(dct_env) > 0:
      # update default config with environment
      self.P("Updating model config with 'SERVING_ENVIRONMENT' config...")
      dct_cfg = self._merge_prepare_config(delta_config=self._environment_variables)
    else:
      self.P("No 'SERVING_ENVIRONMENT' config update.")
    # now update dct_cfg with upstream
    self.P("Updating model config with upstream config...")
    self.config_model = self._merge_prepare_config(
      default_config=dct_cfg,
      delta_config=self._upstream_config,
    )

    self.setup_config_and_validate(self.config_model)

    return
  # END: SUBCLASS CAN OVERWRITE

  def validate_custom_downloadable_model(self):
    if self.cfg_custom_downloadable_model:
      if self.cfg_custom_downloadable_model_url is None:
        self.add_error("Attempted to create a CUSTOM_DOWNLOADABLE_MODEL without a definition URL")

    return
"""

"""